"""
This module contains a collections of the inverse of `privacy_calibrator'.  Given a fixed randomized algorithm and a
desired parameter `delta` it calculates the corresponding (eps,delta)-DP guarantee.

These are building blocks of many differential privacy algorithms.

In some cases, given a fixed randomized algorithm on a fixed data set, it calculates the corresponding (eps,delta)-pDP.
"""

import numpy as np
from autodp import dp_acct, rdp_acct, rdp_bank, utils
from scipy.special import comb
from scipy.stats import norm
from scipy.optimize import minimize_scalar, root_scalar


def get_eps_rdp(func, delta):
    """
    This is the generic function that uses RDP accountant and RDP function to solve for eps given delta
    :param func:
    :param delta:
    :return: The corresponding epsilon
    """
    assert (delta >= 0)
    acct = rdp_acct.anaRDPacct(m=10, m_max=10)
    acct.compose_mechanism(func)
    return acct.get_eps(delta)


def get_eps_rdp_subsampled(func, delta, prob):
    """
    This is the generic function that uses RDP accountant and RDP function to solve for eps given delta
    :param func:
    :param delta:
    :return: The corresponding epsilon
    """
    assert (delta >= 0)
    assert (prob >= 0)
    if prob == 0:
        return 0
    elif prob == 1:
        return get_eps_rdp(func, delta)
    else:
        acct = rdp_acct.anaRDPacct()
        acct.compose_subsampled_mechanism(func, prob)
        return acct.get_eps(delta)


# Get the eps and delta for a single Gaussian mechanism
def get_eps_gaussian(sigma, delta):
    """ This function calculates the eps for Gaussian Mech given sigma and delta"""
    assert (delta >= 0)
    func = lambda x: rdp_bank.RDP_gaussian({'sigma': sigma}, x)
    return get_eps_rdp(func, delta)


def get_logdelta_ana_gaussian(sigma, eps):
    """ This function calculates the delta parameter for analytical gaussian mechanism given eps"""
    assert (eps >= 0)
    s, mag = utils.stable_log_diff_exp(norm.logcdf(0.5 / sigma - eps * sigma),
                                       eps + norm.logcdf(-0.5 / sigma - eps * sigma))
    return mag


def get_eps_ana_gaussian(sigma, delta):
    """ This function calculates the gaussian mechanism given sigma and delta using analytical GM"""
    # Basically inverting the above function by solving a nonlinear equation
    assert (delta >= 0 and delta <= 1)

    if delta == 0:
        return np.inf
    if np.log(delta) >= get_logdelta_ana_gaussian(sigma, 0.0):
        return 0.0

    def fun(x):
        if x < 0:
            return np.inf
        else:
            return get_logdelta_ana_gaussian(sigma, x) - np.log(delta)

    eps_upperbound = 1 / 2 / sigma ** 2 + 1 / sigma * np.sqrt(2 * np.log(1 / delta))
    results = root_scalar(fun, bracket=[0, eps_upperbound])
    if results.converged:
        return results.root
    else:
        raise RuntimeError(f"Failed to find epsilon: {results.flag}")


# def get_eps_gaussian_svt(x, sigma, delta, k, c, c_tilde):
def get_eps_gaussian_svt(params, delta):
    """
    Implements Theorem 12 (stage-wise length-capped Gaussian SVT) in Zhu et.al., NeurIPS-20.

    we want to partition c into [c/c'] parts, each part using (k choose c')
    need to check whether (k choose c') > log(1/delta')
    k is the maximum number of queries to answer for each chunk
    x is delta for each chunk, it needs to be negative
    :param x:
    :param sigma:
    :param delta:
    :return:
    """
    # {'sigma': params['sigma'], 'k': params['k'], 'c': params['c']}
    sigma = params['sigma']
    k = params.get('k', 1)
    c = params.get('c', 1)
    c_tilde = params.get('c_tilde', int(np.sqrt(c)))  # default c_tilde is sqrt(c)
    coeff = comb(k, c_tilde)

    # the function below returns the composed epsilon if each chunk is assigned a e^x budget of delta (the delta'
    # used in Theorem 12).
    def search_per_delta_each_chunk(x, sigma, c, c_tilde, k, coeff):
        acct = dp_acct.DP_acct()
        per_delta = np.exp(x)  # per_delta for each c' chunk
        assert per_delta < 1.0 / (coeff)
        # compute the eps per step with 1/(sigma_1**2) + sqrt(2/simga_1**2 *(log k + log(1/epr_delta)))
        # compose eps for each chunk
        while c:
            if c >= c_tilde:
                c = c - c_tilde
            else:
                # the remaining part c // c_tilde
                c = 0
                c_tilde = c
                coeff = comb(k, c_tilde)
            per_eps = (1.0 + c_tilde) / (2 * sigma ** 2) + np.sqrt(
                (1.0 + c_tilde) / (2 * sigma ** 2) * (np.log(coeff) - x))
            acct.update_DPlosses(per_eps, per_delta)
        compose_eps = acct.get_eps(delta)
        return compose_eps  # composed epsilon over c/c' chunks.

    fun = lambda x: search_per_delta_each_chunk(x, sigma, c, c_tilde, k, coeff)
    # the bound of per chunk delta (delta').
    const = 10
    right_bound = min(delta / (c_tilde + 1), 1.0 / coeff)
    left_bound =delta / (c_tilde * 10)
    # ensures the left bound is smaller than the right bound
    while 10 * left_bound > right_bound:
        const = const * 10
        left_bound = left_bound * 0.1
    results = minimize_scalar(fun, method='Bounded', bounds=[np.log(left_bound), np.log(right_bound)], options={'disp': False})
    return results.fun


def get_eps_laplace(b, delta):
    assert (delta >= 0)
    func = lambda x: rdp_bank.RDP_laplace({'b': b}, x)
    return get_eps_rdp(func, delta)


def get_eps_randresp(p, delta):
    assert (delta >= 0)
    func = lambda x: rdp_bank.RDP_randresponse({'p': p}, x)
    return get_eps_rdp(func, delta)


def get_eps_randresp_optimal(p, delta):
    assert (delta >= 0)
    if p < 0.5:
        p = 1 - p

    if p == 0 or p == 1:
        return np.inf

    eps_max = np.log(p) - np.log(1 - p)
    if delta == 0:
        return eps_max
    elif delta >= 2 * p - 1:
        return 0.0
    else:
        return np.log(p - delta) - np.log(1 - p)



"""
This file contains the implementation of the main class object:  anaRDPacct --- an analytical moment accountant
that keeps track the effects of a hetereogeneous sequence of randomized algorithms using the RDP technique.

In particular it supports amplification of RDP by subsampling without replacement and the amplification of RDP
by poisson sampling, but unfortunately not (yet) together.


* The RDP accountant API is deprecated.
* All functionality of an RDP accountant can be achieved via the Mechanism API.

"""






import numpy as np
from scipy.optimize import minimize_scalar
import sys
sys.path.append('..')
import autodp

from autodp import utils, rdp_bank
from autodp.privacy_calibrator import subsample_epsdelta
import scipy
import math

def general_upperbound(func, mm, prob):
    """

    :param func:
    :param mm: alpha in RDP
    :param prob: sample probability
    :return: the upperbound in theorem 1 in 2019 ICML,could be applied for general case(including poisson distribution)
    k_approx = 100 k approximation is applied here
    """
    def cgf(x):
        return (x - 1) * func(x)

    if np.isinf(func(mm)):
        return np.inf
    if mm == 1 or mm == 0:
        return 0

    cur_k = np.minimum(50, mm - 1) # choose small k-approx for general upperbound (here is 50) in case of scipy-accuracy
    log_term_1 = mm * np.log(1 - prob)
    #logBin = utils.get_binom_coeffs(mm)
    log_term_2 = np.log(3) - func(mm) + mm * utils.stable_logsumexp_two(np.log(1 - prob), np.log(prob) + func(mm))
    neg_term_3 = [np.log(scipy.special.comb(mm,l)) + np.log(3) + (mm - l) * np.log(1 - prob) + l * np.log(prob) +
                  utils.stable_log_diff_exp((l - 1) * func(mm), cgf(l))[1] for l in
                  range(3, cur_k + 1)]
    neg_term_4 = np.log(mm*(mm - 1)/2) + 2 * np.log(prob) + (mm - 2) * np.log(
        1 - prob) + utils.stable_log_diff_exp(np.log(3) + func(mm), func(2))[1]
    neg_term_5 = np.log(2) + np.log(prob) + np.log(mm) + (mm - 1) * np.log(1 - prob)
    neg_term_6 = mm * np.log(1 - prob) + np.log(3) - func(mm)
    pos_term = utils.stable_logsumexp([log_term_1, log_term_2])
    neg_term_3.append(neg_term_4)
    neg_term_3.append(neg_term_5)
    neg_term_3.append(neg_term_6)
    neg_term = utils.stable_logsumexp(neg_term_3)
    bound = utils.stable_log_diff_exp(pos_term, neg_term)[1]
    return bound

def fast_subsampled_cgf_upperbound(func, mm, prob, deltas_local):
    # evaulate the fast CGF bound for the subsampled mechanism
    # func evaluates the RDP of the base mechanism
    # mm is alpha.  NOT lambda.
    return np.inf

    if np.isinf(func(mm)):
        return np.inf
    if mm == 1:
        return 0

    secondterm = 2 * np.log(prob) + + np.log(mm) + np.log(mm - 1) - np.log(2) \
                  + np.mininum(np.log(4) + func(2.0) + np.log(1 - np.exp(-func(2.0))),
                               func(2.0) + np.mininum(np.log(2),
                                                      2 * (eps_inf + np.log(1 - np.exp(-eps_inf)))))

    # secondterm = np.minimum(np.minimum((2) * np.log(np.exp(func(np.inf)) - 1)
    #                                    + np.minimum(func(2), np.log(4)),
    #                                    np.log(2) + func(2)),
    #                         np.log(4) + 0.5 * deltas_local[int(2 * np.floor(2 / 2.0)) - 1]
    #                         + 0.5 * deltas_local[int(2 * np.ceil(2 / 2.0)) - 1]
    #                         ) + 2 * np.log(prob) + np.log(mm) + np.log(mm - 1) - np.log(2)

    if mm == 2:
        return utils.stable_logsumexp([0, secondterm])

    # approximate the remaining terms using a geometric series or binomial series

    log_exp_eps_minus_one = func(np.inf) + np.log(1 - np.exp(-func(np.inf)))


    if mm == 3:
        return utils.stable_logsumexp([0, secondterm, (3 * (np.log(prob) + np.log(mm))
                                                       + 2*func(mm)
                                                       + np.minumum(np.log(2),
                                                                    3 * log_exp_eps_minus_one))])

    logratio1 = np.log(prob) + np.log(mm) + func(mm)
    logratio2 = logratio1 + log_exp_eps_minus_one

    s, mag = utils.stable_log_diff_exp(1,logratio1)
    s, mag2 = utils.stable_log_diff_exp(1, (mm-3)*logratio1)
    remaining_terms1 = (np.log(2) + 3 * (np.log(prob) + np.log(mm)) + 2*func(mm)
                        + mag2 - mag)

    s, mag = utils.stable_log_diff_exp(1,logratio2)
    s, mag2 = utils.stable_log_diff_exp(1, (mm-3)*logratio2)

    remaining_terms2 = (3 * (np.log(prob) + np.log(mm) + log_exp_eps_minus_one) + 2*func(mm)
                        + mag2 - mag)

    return utils.stable_logsumexp([0, secondterm, np.minimum(remaining_terms1,remaining_terms2)])

    # logratio = np.minimum(logratio1, logratio2)
    # if logratio1 > logratio2:
    #     coeff = 1
    # else:
    #     coeff = 2
    #
    #
    # if mm == 3:
    #     return utils.stable_logsumexp([0, secondterm, np.log(coeff) + 3 * logratio])
    #
    # # Calculate the sum of the geometric series starting from the third term. This is a total of mm-2 terms.
    # if logratio < 0:
    #     geometric_series_bound = np.log(coeff) + 3 * logratio - np.log(1 - np.exp(logratio)) \
    #                              + np.log(1 - np.exp((mm - 2) * logratio))
    # elif logratio > 0:
    #     geometric_series_bound = np.log(coeff) + 3 * logratio + (mm-2) * logratio - np.log(np.exp(logratio) - 1)
    # else:
    #     geometric_series_bound = np.log(coeff) + np.log(mm - 2)
    #
    # # we will approximate using (1+h)^mm
    # logh1 = np.log(prob) + func(mm - 1)
    #
    # logh2 = logh1 + np.log(np.exp(func(np.inf)) - 1)
    #
    # binomial_series_bound1 = np.log(2) + mm * utils.stable_logsumexp_two(0, logh1)
    # binomial_series_bound2 = mm * utils.stable_logsumexp_two(0, logh2)
    #
    # tmpsign, binomial_series_bound1 \
    #     = utils.stable_sum_signed(True, binomial_series_bound1, False, np.log(2)
    #                               + utils.stable_logsumexp([0, logh1 + np.log(mm), 2 * logh1 + np.log(mm)
    #                                                         + np.log(mm - 1) - np.log(2)]))
    # tmpsign, binomial_series_bound2 \
    #     = utils.stable_sum_signed(True, binomial_series_bound2, False,
    #                               utils.stable_logsumexp([0, logh2 + np.log(mm), 2 * logh2 + np.log(mm)
    #                                                       + np.log(mm - 1) - np.log(2)]))
    #
    # remainder = np.min([geometric_series_bound, binomial_series_bound1, binomial_series_bound2])
    #
    # return utils.stable_logsumexp([0, secondterm, remainder])



def fast_poission_subsampled_cgf_upperbound(func, mm, prob):
    # evaulate the fast CGF bound for the subsampled mechanism
    # func evaluates the RDP of the base mechanism
    # mm is alpha.  NOT lambda.

    if np.isinf(func(mm)):
        return np.inf
    if mm == 1:
        return 0
    # Bound #1:   log [ (1-\gamma + \gamma e^{func(mm)})^mm ]
    bound1  = mm * utils.stable_logsumexp_two(np.log(1-prob), np.log(prob) + func(mm))

    # Bound #2:   log [  (1-gamma)^alpha E  [ 1 +  gamma/(1-gamma) E[p/q]]^mm ]
    #     log[  (1-gamma)^\alpha    { 1 + alpha gamma / (1-gamma)  + gamma^2 /(1-gamma)^2 * alpha(alpha-1) /2 e^eps(2))
    #  + alpha \choose 3 * gamma^3 / (1-gamma)^3  / e^(-2 eps(alpha)) * (1 + gamma /(1-gamma) e^{eps(alpha)}) ^ (alpha - 3) }
    # ]
    if mm >= 3:
        bound2 = utils.stable_logsumexp([mm * np.log(1-prob), (mm-1) * np.log(1-prob) + np.log(mm) + np.log(prob),
                                     (mm-2)*np.log(1-prob) + 2 * np.log(prob) + np.log(mm) + np.log(mm-1) + func(2),
                                     np.log(mm) + np.log(mm-1) + np.log(mm-2) - np.log(3*2) + 3 * np.log(prob)
                                     + (mm-3)*np.log(1-prob) + 2 * func(mm) +
                                     (mm-3) * utils.stable_logsumexp_two(0, np.log(prob) - np.log(1-prob) + func(mm))])
    else:
        bound2 = bound1

    #print('www={} func={} mm={}'.format(np.exp(func(mm))-1),func, mm)
    #print('bound1 ={} bound2 ={}'.format(bound1,bound2))
    return np.minimum(bound1,bound2)

def fast_k_subsample_upperbound(func, mm, prob, k):
    """

     :param func:
     :param mm:
     :param prob: sample probability
     :param k: approximate term
     :return: k-term approximate upper bound in therorem 11 in ICML-19
     """
    def cgf(x):
        return (x - 1) * func(x)

    if np.isinf(func(mm)):
        return np.inf
    if mm == 1:
        return 0
    #logBin = utils.get_binom_coeffs(mm)
    cur_k = np.minimum(k, mm - 1)
    if (2 * cur_k) >= mm:
        exact_term_1 = (mm - 1) * np.log(1 - prob) + np.log(mm * prob - prob + 1)
        exact_term_2 = [np.log(scipy.special.comb(mm,l))  + (mm - l) * np.log(1 - prob) + l * np.log(prob) + cgf(l) for l in
                        range(2, mm + 1)]
        exact_term_2.append(exact_term_1)
        bound = utils.stable_logsumexp(exact_term_2)
        return bound

    s, mag1 = utils.stable_log_diff_exp(0, -func(mm - cur_k))
    new_log_term_1 = np.log(1 - prob) * mm + mag1
    new_log_term_2 = -func(mm - cur_k) + mm * utils.stable_logsumexp_two(np.log(1 - prob),
                                                                         np.log(prob) + func(mm - cur_k))
    new_log_term_3 = [np.log(scipy.special.comb(mm,l)) + (mm - l) * np.log(1 - prob) + l * np.log(prob) +
                      utils.stable_log_diff_exp((l - 1) * func(mm - cur_k), cgf(l))[1] for l in
                      range(2, cur_k + 1)]
    if len(new_log_term_3) > 0:
        new_log_term_3 = utils.stable_logsumexp(new_log_term_3)
    else:
        return utils.stable_logsumexp_two(new_log_term_1, new_log_term_2)
    new_log_term_4 = [np.log(scipy.special.comb(mm,mm-l)) + (mm - l) * np.log(1 - prob) + l * np.log(prob) +
                      utils.stable_log_diff_exp(cgf(l), (l - 1) * func(mm - cur_k))[1] for l in
                      range(mm - cur_k + 1, mm + 1)]
    new_log_term_4.append(new_log_term_1)
    new_log_term_4.append(new_log_term_2)
    new_log_term_4 = utils.stable_logsumexp(new_log_term_4)
    s, new_log_term_5 = utils.stable_log_diff_exp(new_log_term_4, new_log_term_3)
    new_bound = new_log_term_5
    return new_bound


class anaRDPacct:
    """A class that keeps track of the analytical expression of the RDP --- 1/(alpha-1)*CGF of the privacy loss R.V."""
    def __init__(self, m=100, tol=0.1, m_max=500, m_lin_max=10000, approx = False, verbose=False):
        # m_max indicates the number that we calculate binomial coefficients exactly up to.
        # beyond that we use Stirling approximation.

        # ------ Class Attributes -----------
        self.m = m # default number of binomial coefficients to precompute
        self.m_max = m_max # An upper bound of the quadratic dependence
        self.m_lin_max = m_lin_max # An upper bound of the linear dependence.
        self.verbose = verbose
        self.approx = approx # If true, use the fast k-term approximation
        self.lambs = np.linspace(1, self.m, self.m).astype(int) # Corresponds to \alpha = 2,3,4,5,.... for RDP

        self.alphas = np.linspace(1, self.m, self.m).astype(int)
        self.RDPs_int = np.zeros_like(self.alphas, float)

        self.n=0
        self.RDPs = [] # analytical CGFs
        self.coeffs = []
        self.RDP_inf = .0 # This is effectively for pure DP.
        self.logBinomC = utils.get_binom_coeffs(self.m + 1)  # The logBinomC is only needed for subsampling mechanisms.
        self.idxhash = {} # save the index of previously used algorithms
        self.cache = {} # dictionary to save results from previously seen algorithms
        self.deltas_cache = {} # dictionary to save results of all discrete derivative path
        self.evalRDP = lambda x: 0
        self.flag = True # a flag indicating whether evalCGF is out of date
        self.flag_subsample = False # a flag to indicate whether we need to expand the logBinomC.
        self.tol = tol


    # ---------- Methods ------------
    def build_zeroth_oracle(self):
        self.evalRDP = lambda x:  sum([c * item(x) for (c, item) in zip(self.coeffs, self.RDPs)])

    def plot_rdp(self):
        if not self.flag:
            self.build_zeroth_oracle()
            self.flag = True

        import matplotlib.pyplot as plt
        plt.figure(num=None, figsize=(12, 8), dpi=80, facecolor='w', edgecolor='k')
        x = range(0,self.m,1)
        y = [self.evalRDP(item) for item in x]
        plt.loglog(x, y)
        plt.show()


    def plot_cgf_int(self):
        import matplotlib.pyplot as plt
        plt.figure(num=None, figsize=(12, 8), dpi=80, facecolor='w', edgecolor='k')
        plt.plot(self.alphas, self.RDPs_int)
        plt.xlabel(r'$\lambda$')
        plt.ylabel('CGF')
        plt.show()

    def plot_rdp_int(self):
        import matplotlib.pyplot as plt
        plt.figure(num=None, figsize=(12, 8), dpi=80, facecolor='w', edgecolor='k')
        plt.loglog(self.alphas, self.RDPs_int)
        if not self.flag:
            self.build_zeroth_oracle()
            self.flag = True
        x = range(1,self.m_lin_max,1)
        y = [self.evalRDP(item) for item in x]
        plt.loglog(x, y)
        plt.xlabel(r'$\alpha$')
        plt.ylabel(r'RDP $\epsilon$')
        plt.show()


    def get_rdp(self,alphas):
        # alphas is a numpy array or a list of numbers
        # we will return a numpy array of the corresponding RDP
        if not self.flag:
            self.build_zeroth_oracle()
            self.flag = True
        alphas = np.array(alphas)
        assert(np.all(alphas >= 1))
        rdp_list = []
        for alpha in alphas:
            rdp_list.append(self.evalRDP(alpha))

        return np.array(rdp_list)

    def get_eps(self, delta): # minimize over \lambda
        if not self.flag:
            self.build_zeroth_oracle()
            self.flag = True

        if delta<0 or delta > 1:
            print("Error! delta is a probability and must be between 0 and 1")
        if delta == 0:
            return self.RDP_inf
        else:
            def fun(x): # the input the RDP's \alpha
                if x <= 1:
                    return np.inf
                else:
                    return np.log(1 / delta)/(x-1) + self.evalRDP(x)

            def fun_int(i): # the input is RDP's \alpha in integer
                if i <= 1 | i >= len(self.RDPs_int):
                    return np.inf
                else:
                    return np.log(1 / delta) / (i-1) + self.RDPs_int[i - 1]


            # When do we have computational constraints?
            # Only when we have subsampled items.

            # First check if the forward difference is positive at self.m, or if it is infinite
            while (self.m<self.m_max) and (not np.isposinf(fun(self.m))) and (fun_int(self.m-1)-fun_int(self.m-2) < 0):
                # If so, double m, expand logBimomC until the forward difference is positive


                if self.flag_subsample:

                    # The following line is m^2 time.
                    self.logBinomC = utils.get_binom_coeffs(self.m*2+1)

                    # Update deltas_caches
                    for key, val in self.deltas_cache.items():
                        if type(key) is tuple:
                            func_tmp = key[0]
                        else:
                            func_tmp = key
                        cgf = lambda x:  x*func_tmp(x+1)
                        deltas,signs_deltas = utils.get_forward_diffs(cgf,self.m*2)

                        self.deltas_cache[key] = [deltas, signs_deltas]

                new_alphas = range(self.m + 1, self.m * 2 + 1, 1)
                self.alphas = np.concatenate((self.alphas, np.array(new_alphas)))  # array of integers
                self.m = self.m * 2

            mm = np.max(self.alphas)

            rdp_int_new = np.zeros_like(self.alphas, float)

            for key,val in self.cache.items():
                idx = self.idxhash[key]
                rdp = self.RDPs[idx]
                newarray = np.zeros_like(self.alphas, float)
                for j in range(2,mm+1,1):
                    newarray[j-1] = rdp(1.0*j)
                newarray[0]=newarray[1]
                coeff = self.coeffs[idx]
                rdp_int_new += newarray * coeff
                self.cache[key] = newarray

            self.RDPs_int = rdp_int_new

            # # update the integer CGF and the cache for each function
            # rdp_int_new = np.zeros_like(self.RDPs_int)
            # for key,val in self.cache.items():
            #     idx = self.idxhash[key]
            #     rdp = self.RDPs[idx]
            #     newarray = np.zeros_like(self.RDPs_int)
            #     for j in range(self.m):
            #         newarray[j] = rdp(1.0*(j+self.m+1))
            #
            #     coeff = self.coeffs[idx]
            #     rdp_int_new += newarray * coeff
            #     self.cache[key] = np.concatenate((val, newarray))
            #
            # # update the corresponding quantities
            # self.RDPs_int = np.concatenate((self.RDPs_int, rdp_int_new))

            #self.m = self.m*2

            bestint = np.argmin(np.log(1 / delta)/(self.alphas[1:]-1) + self.RDPs_int[1:]) + 1

            if bestint == self.m-1:
                if self.verbose:
                    print('Warning: Reach quadratic upper bound: m_max.')
                # In this case, we matches the maximum qudaratic upper bound
                # Fix it by calling O(1) upper bounds and do logarithmic search
                cur = fun(bestint)
                while (not np.isposinf(cur)) and fun(bestint-1)-fun(bestint-2) < -1e-8:
                    bestint = bestint*2
                    cur = fun(bestint)
                    if bestint > self.m_lin_max and self.approx ==True:
                        print('Warning: Reach linear upper bound: m_lin_max.')
                        return cur

                results = minimize_scalar(fun, method='Bounded', bounds=[self.m-1, bestint + 2],
                                          options={'disp': False})
                if results.success:
                    return results.fun
                else:
                    raise RuntimeError(f"Optimal RDP order not found: {results.message}")
                #return fun(bestint)

            if bestint == 0:
                if self.verbose:
                    print('Warning: Smallest alpha = 1.')

            # find the best integer alpha.
            bestalpha = self.alphas[bestint]

            results = minimize_scalar(fun,  method='Bounded',bounds=[bestalpha-1, bestalpha+1],
                                      options={'disp':False})
            # the while loop above ensures that bestint+2 is at most m, and also bestint is at least 0.
            if results.success:
                return results.fun
            else:
                # There are cases when certain \delta is not feasible.
                # For example, let p and q be uniform the privacy R.V. is either 0 or \infty and unless all \infty
                # events are taken cared of by \delta, \epsilon cannot be < \infty
                return np.inf

    def compose_mechanism(self, func, coeff=1.0):
        self.flag = False
        if func in self.idxhash:
            self.coeffs[self.idxhash[func]] += coeff
            # also update the integer CGFs
            self.RDPs_int += self.cache[func] * coeff

        else:
            # book keeping
            self.idxhash[func] = self.n
            self.n += 1
            self.coeffs.append(coeff)
            # update the analytical
            self.RDPs.append(func)

            # also update the integer results
            if func in self.cache:
                tmp = self.cache[func]
            else:
                tmp = np.zeros_like(self.RDPs_int, float)
                for i in range(self.m):
                    tmp[i] = func(i+1)
                self.cache[func] = tmp  # save in cache
            self.RDPs_int += tmp * coeff

        self.RDP_inf += func(np.inf) * coeff
    #795010
    #imple 100
    def compose_subsampled_mechanism(self, func, prob, coeff=1.0, improved_bound_flag=False):
        """
            # This function is for subsample without replacements
        :param func:  RDP function of the mechanism before amplification by sampling
        :param prob:  proportion of the data to sample
        :param coeff: number of times the subsampled mechanism is being composed.
        :param improved_bound_flag:
            - If True, then it uses Theorem 27 of https://arxiv.org/pdf/1808.00087.pdf
            - If False (default value), it uses Theorem 9 of https://arxiv.org/pdf/1808.00087.pdf
            To qualify for the improved bound, the mechanism needs to have a pair of neighboring
            datasets that is worst for all Renyi-divergence and Pearson-Vajda divergence;
            Also, the RDP bound needs to be tight (see Definition 26 from the same paper).
            Gaussian mechanism, Laplace mechanism and many others satisfy this condition.

        :return:  nothing  (updates to the RDP accountant's attribute)
        """

        # (find a random subset of proportion prob)
        self.flag = False
        self.flag_subsample = True
        if (func, prob) in self.idxhash:
            idx = self.idxhash[(func, prob)]
            # update the coefficients of each function
            self.coeffs[idx] += coeff
            # also update the integer CGFs
            self.RDPs_int += self.cache[(func, prob)] * coeff
        else:
            def cgf(x):
                return x * func(x+1)

            if not improved_bound_flag:
                def subsample_func_int(x):
                    # output the cgf of the subsampled mechanism
                    mm = int(x)
                    eps_inf = func(np.inf)

                    moments_two = 2 * np.log(prob) + utils.logcomb(mm,2) \
                                  + np.minimum(np.log(4) + func(2.0) + np.log(1-np.exp(-func(2.0))),
                                               func(2.0) + np.minimum(np.log(2),
                                                            2 * (eps_inf+np.log(1-np.exp(-eps_inf)))))
                    moment_bound = lambda j: np.minimum(j * (eps_inf + np.log(1-np.exp(-eps_inf))),
                                                        np.log(2)) + cgf(j - 1) \
                                             + j * np.log(prob) + utils.logcomb(mm, j)
                    moments = [moment_bound(j) for j in range(3, mm + 1, 1)]
                    return np.minimum((x-1)*func(x), utils.stable_logsumexp([0,moments_two] + moments))
            else:
                # we need forward differences of exp(cgf)
                # The following line is the numerically stable way of implementing it.
                # The output is in polar form with logarithmic magnitude
                deltas, signs_deltas = utils.get_forward_diffs(cgf, self.m)

                #deltas1, signs_deltas1 = get_forward_diffs_direct(func, self.m)

                #tmp = deltas-deltas1

                self.deltas_cache[(func,prob)] = [deltas,signs_deltas]

                def subsample_func_int(x):
                    # This function evaluates teh CGF at alpha = x, i.e., lamb =  x- 1
                    deltas_local, signs_deltas_local = self.deltas_cache[(func,prob)]
                    if np.isinf(func(x)):
                        return np.inf

                    mm = int(x)
                    eps_inf = func(np.inf)

                    moments_two = 2 * np.log(prob) + utils.logcomb(mm, 2) \
                                  + np.minimum(
                        np.log(4) + func(2.0) + np.log(1 - np.exp(-func(2.0))),
                        func(2.0) + np.minimum(np.log(2),
                                               2 * (eps_inf + np.log(1 - np.exp(-eps_inf)))))

                    moment_bound = lambda j: np.minimum(np.log(4) + 0.5*deltas_local[int(2*np.floor(j/2.0))-1]
                                                        + 0.5*deltas_local[int(2*np.ceil(j/2.0))-1],
                                                        np.minimum(j * (eps_inf + np.log(1 - np.exp(-eps_inf))),
                                                                   np.log(2))
                                                        + cgf(j - 1)) \
                                             + j * np.log(prob) + utils.logcomb(mm, j)

                    moment_bound_linear = lambda j: np.minimum(j * (eps_inf + np.log(1-np.exp(-eps_inf))),
                                                        np.log(2)) + cgf(j - 1) \
                                             + j * np.log(prob) + utils.logcomb(mm, j)

                    fastupperbound = fast_subsampled_cgf_upperbound(func, mm, prob, deltas_local)

                    if mm <= self.alphas[-1]: # compute the bound exactly. Requires book keeping of O(x^2)
                        #
                        # moments = [ np.minimum(np.minimum((j)*np.log(np.exp(func(np.inf))-1) + np.minimum(cgf(j-1),np.log(4)),
                        #                                   np.log(2) + cgf(j-1)),
                        #                        np.log(4) + 0.5*deltas_local[int(2*np.floor(j/2.0))-1]
                        #                        + 0.5*deltas_local[int(2*np.ceil(j/2.0))-1]) + j*np.log(prob)
                        #             +self.logBinomC[int(mm), j] for j in range(2,int(mm+1),1)]
                        moments = [moment_bound(j) for j in range(3, mm + 1, 1)]

                        return np.minimum(fastupperbound, utils.stable_logsumexp([0, moments_two] + moments))
                    elif mm <= self.m_lin_max:  # compute the bound with stirling approximation. Everything is O(x) now.
                        # moment_bound = lambda j: np.minimum(j * np.log(np.exp(func(np.inf)) - 1)
                        #                                     + np.minimum(cgf(j - 1), np.log(4)), np.log(2)
                        #                                     + cgf(j - 1)) + j * np.log(prob) + utils.logcomb(mm, j)
                        # moments = [moment_bound(j) for j in range(2,mm+1,1)]


                        moments = [moment_bound_linear(j) for j in range(3, mm + 1, 1)]

                        return np.minimum(fastupperbound, utils.stable_logsumexp([0, moments_two] + moments))
                    else: # Compute the O(1) upper bound
                        return fastupperbound



            def subsample_func(x):
                # This function returns the RDP at alpha = x
                # RDP with the linear interpolation upper bound of the CGF

                epsinf, tmp = subsample_epsdelta(func(np.inf),0,prob)

                if np.isinf(x):
                    return epsinf
                if prob == 1.0:
                    return func(x)

                if (x >= 1.0) and (x <= 2.0):
                    return np.minimum(epsinf, subsample_func_int(2.0) / (2.0-1))
                if np.equal(np.mod(x, 1), 0):
                    return np.minimum(epsinf, subsample_func_int(x) / (x-1) )
                xc = math.ceil(x)
                xf = math.floor(x)
                return np.min(
                    [epsinf,func(x),
                     ((x-xf)*subsample_func_int(xc) + (1-(x-xf))*subsample_func_int(xf)) / (x-1)]
                )


            # book keeping
            self.idxhash[(func, prob)] = self.n # save the index
            self.n += 1 # increment the number of unique mechanisms
            self.coeffs.append(coeff) # Update the coefficient
            self.RDPs.append(subsample_func) # update the analytical functions

            # also update the integer results up to m_max.
            if (func,prob) in self.cache:
                results = self.cache[(func,prob)]
            else:
                results = np.zeros_like(self.RDPs_int, float)
                # m = np.max(self.lambs)
                mm = np.max(self.alphas)
                for alpha in range(2, mm+1):
                    results[alpha-1] = subsample_func(alpha)
                results[0] = results[1] # Provide the trivial upper bound of RDP at alpha = 1 --- the KL privacy.
                self.cache[(func,prob)] = results # save in cache

            self.RDPs_int += results * coeff
        # update the pure DP
        eps, delta = subsample_epsdelta(func(np.inf), 0, prob)
        self.RDP_inf += eps * coeff

    def compose_subsampled_mechanisms_lowerbound(self, func, prob, coeff=1.0):
        self.compose_poisson_subsampled_mechanisms(func, prob, coeff=coeff)

    def compose_poisson_subsampled_mechanisms(self, func, prob, coeff=1.0):
        # This function implements the lower bound for subsampled RDP.
        # It is also the exact formula of poission_subsampled RDP for many mechanisms including Gaussian mech.
        #
        # At the moment, we do not support mixing poisson subsampling and standard subsampling.
        # TODO: modify the caching identifies so that we can distinguish different types of subsampling
        #
        self.flag = False
        self.flag_subsample = True
        if (func, prob) in self.idxhash:
            idx = self.idxhash[(func, prob)] # TODO: this is really where it needs to be changed.
            # update the coefficients of each function
            self.coeffs[idx] += coeff
            # also update the integer CGFs
            self.RDPs_int += self.cache[(func, prob)] * coeff
        else: # compute an easy to compute upper bound of it.

            def cgf(x):
                return x * func(x+1)

            def subsample_func_int(x):
                # This function evaluates teh CGF at alpha = x, i.e., lamb =  x- 1

                if np.isinf(func(x)):
                    return np.inf

                mm = int(x)
                #
                fastbound = fast_poission_subsampled_cgf_upperbound(func, mm, prob)

                k = self.alphas[-1]
                fastbound_k = fast_k_subsample_upperbound(func, mm, prob,k)
                if self.approx == True:
                    return fastbound_k
                #fastbound = min(fastbound, fastbound_k)
                if x <= self.alphas[-1]: # compute the bound exactly.
                    moments = [cgf(j-1) +j*np.log(prob) + (mm-j) * np.log(1-prob)
                               + self.logBinomC[mm, j] for j in range(2,mm+1,1)]

                    return utils.stable_logsumexp([(mm-1)*np.log(1-prob)+np.log(1+(mm-1)*prob)]+moments)
                elif mm <= self.m_lin_max:
                    moments = [cgf(j-1) +j*np.log(prob) + (mm-j) * np.log(1-prob)
                               + utils.logcomb(mm,j) for j in range(2,mm+1,1)]
                    return utils.stable_logsumexp([(mm-1)*np.log(1-prob)+np.log(1+(mm-1)*prob)] + moments)
                else:
                    return fastbound

            def subsample_func(x): # linear interpolation upper bound
                # This function implements the RDP at alpha = x

                if np.isinf(func(x)):
                    return np.inf
                if prob == 1.0:
                    return func(x)
                if prob == 0:
                    return 0

                epsinf, tmp = subsample_epsdelta(func(np.inf),0,prob)

                if np.isinf(x):
                    return epsinf
                if (x >= 1.0) and (x <= 2.0):
                    return np.minimum(epsinf, subsample_func_int(2.0) / (2.0-1))
                if np.equal(np.mod(x, 1), 0):
                    return np.minimum(epsinf, subsample_func_int(x) / (x-1) )
                xc = math.ceil(x)
                xf = math.floor(x)
                return np.minimum(
                    epsinf,
                    ((x-xf)*subsample_func_int(xc) + (1-(x-xf))*subsample_func_int(xf)) / (x-1)
                )

            # book keeping
            self.idxhash[(func, prob)] = self.n # save the index
            self.n += 1 # increment the number of unique mechanisms
            self.coeffs.append(coeff) # Update the coefficient
            self.RDPs.append(subsample_func) # update the analytical functions

            # also update the integer results, with a vectorized computation.
            # TODO: pre-computing subsampled RDP for integers is error-prone (implement the same thing twice)
            # TODO: and its benefits are not clear. We should consider removing it and simply call the lambda function.
            #
            if (func,prob) in self.cache:
                results = self.cache[(func,prob)]
            else:
                results = np.zeros_like(self.RDPs_int, float)
                mm = np.max(self.alphas)  # evaluate the RDP up to order mm
                jvec = np.arange(2, mm + 1)
                logterm3plus = np.zeros_like(results)  # This saves everything from j=2 to j = m+1
                for j in jvec:
                    logterm3plus[j-2] = cgf(j-1) + j * np.log(prob)  #- np.log(1-prob))

                for alpha in range(2, mm+1):
                    if np.isinf(logterm3plus[alpha-1]):
                        results[alpha-1] = np.inf
                    else:
                        tmp = utils.stable_logsumexp(logterm3plus[0:alpha-1] + self.logBinomC[alpha , 2:(alpha + 1)]
                                               + (alpha+1-jvec[0:alpha-1])*np.log(1-prob))
                        results[alpha-1] = utils.stable_logsumexp_two((alpha-1)*np.log(1-prob)
                                                                    + np.log(1+(alpha-1)*prob), tmp) / (1.0*alpha-1)

                results[0] = results[1]  # Provide the trivial upper bound of RDP at alpha = 1 --- the KL privacy.
                self.cache[(func,prob)] = results # save in cache
            self.RDPs_int += results * coeff
        # update the pure DP tracker
        eps, delta = subsample_epsdelta(func(np.inf), 0, prob)
        self.RDP_inf += eps * coeff


    def compose_poisson_subsampled_mechanisms1(self, func, prob, coeff=1.0):
        # This function implements the general amplification bounds for Poisson sampling.
        # No additional assumptions are needed.

        # At the moment, we do not support mixing poisson subsampling and standard subsampling.
        #
        self.flag = False
        self.flag_subsample = True
        if (func, prob) in self.idxhash:
            idx = self.idxhash[(func, prob)]
            # update the coefficients of each function
            self.coeffs[idx] += coeff
            # also update the integer CGFs
            self.RDPs_int += self.cache[(func, prob)] * coeff
        else: # compute an easy to compute upper bound of it.

            cgf = lambda x:  x*func(x+1)

            def subsample_func_int(x):
                # This function evaluates the CGF at alpha = x, i.e., lamb =  x- 1
                if np.isinf(func(x)):
                    return np.inf
                if prob == 1.0:
                    return func(x)

                mm = int(x)

                fastbound = fast_poission_subsampled_cgf_upperbound(func, mm, prob)

                if x <= self.alphas[-1]: # compute the bound exactly.
                    moments = [cgf(1) + 2*np.log(prob) + (mm-2) * np.log(1 - prob) + self.logBinomC[mm, 2]]
                    moments = moments + [cgf(j-1+1) +j*np.log(prob) + (mm-j) * np.log(1 - prob)
                               + self.logBinomC[mm, j] for j in range(3,mm+1,1)]

                    return utils.stable_logsumexp([(mm-1)*np.log(1-prob)+np.log(1+(mm-1)*prob)]+moments)
                elif mm <= self.m_lin_max:
                    moments = [cgf(1) + 2*np.log(prob) + (mm-2) * np.log(1 - prob) + utils.logcomb(mm, 2)]
                    moments = moments + [cgf(j-1+1) +j*np.log(prob) + (mm-j) * np.log(1 - prob)
                               + utils.logcomb(mm, j) for j in range(3,mm+1,1)]
                    return utils.stable_logsumexp([(mm-1)*np.log(1-prob)+np.log(1+(mm-1)*prob)]+moments)
                else:
                    return fastbound


            def subsample_func(x): # linear interpolation upper bound
                if prob == 0:
                    return 0
                if prob == 1.0:
                    return func(x)
                epsinf, tmp = subsample_epsdelta(func(np.inf),0,prob)

                if np.isinf(x):
                    return epsinf
                if (x >= 1.0) and (x <= 2.0):
                    return np.minimum(epsinf, subsample_func_int(2.0) / (2.0-1))
                if np.equal(np.mod(x, 1), 0):
                    return np.minimum(epsinf, subsample_func_int(x) / (x-1) )
                xc = math.ceil(x)
                xf = math.floor(x)
                return np.minimum(
                    epsinf,
                    ((x-xf)*subsample_func_int(xc) + (1-(x-xf))*subsample_func_int(xf)) / (x-1)
                )

            # book keeping
            self.idxhash[(func, prob)] = self.n # save the index
            self.n += 1 # increment the number of unique mechanisms
            self.coeffs.append(coeff) # Update the coefficient
            self.RDPs.append(subsample_func) # update the analytical functions

            # also update the integer results
            if (func,prob) in self.cache:
                results = self.cache[(func,prob)]
            else:
                results = np.zeros_like(self.RDPs_int, float)
                mm = np.max(self.alphas)  # evaluate the RDP up to order mm

                for alpha in range(2, mm+1):
                    results[alpha-1] = subsample_func_int(alpha)
                results[0] = results[1]  # Provide the trivial upper bound of RDP at alpha = 1 --- the KL privacy.
                self.cache[(func,prob)] = results # save in cache
            self.RDPs_int += results * coeff
        # update the pure DP tracker
        eps, delta = subsample_epsdelta(func(np.inf), 0, prob)
        self.RDP_inf += eps * coeff


# TODO: 1. Modularize the several Poission sampling versions.  2. Support both sampling schemes together.



"""
Example of a specific transformer that outputs the composition
1. class Composition: supports RDP-based composition.
2. class ComposeAFA: supports characteristic function (phi-function) based composition.
3. class ComposeGaussian: supports composition of only Gaussian mechanisms.
4. class AmplificationBySampling: supports RDP and DP based amplification rule.
5. classAmplificationBySampling_pld: amplification by sampling for privacy loss distribution.

"""

from autodp.autodp_core import Mechanism, Transformer
from copy import deepcopy
import math
import numpy as np

from autodp import mechanism_zoo, rdp_acct, phi_bank


# The generic composition class
class Composition(Transformer):
    """ Composition is a transformer that takes a list of Mechanisms and number of times they appear,
    and output a Mechanism that represents the composed mechanism"""
    def __init__(self):
        Transformer.__init__(self)
        """
        Args:

        """
        self.name = 'Composition'

        # Update the function that is callable
        self.transform = self.compose

    def compose(self, mechanism_list, coeff_list, RDP_compose_only=True, BBGHS_conversion=True, fDP_based_conversion=False):
        # Make sure that the mechanism has a unique list
        # for example, if there are two Gaussian mechanism with two different sigmas, call it
        # Gaussian1, and Gaussian2

        # if RDP_compose_only is true, we use only RDP composition.


        newmech = Mechanism()

        # update the functions
        def newrdp(x):
            return sum([c * mech.RenyiDP(x) for (mech, c) in zip(mechanism_list, coeff_list)])
        newmech.propagate_updates(newrdp, 'RDP',  BBGHS_conversion= BBGHS_conversion, fDP_based_conversion=fDP_based_conversion)

        # TODO: the fDP_based_conversion sometimes fails due to undefined RDP with alpha < 1

        newmech.eps_pureDP = sum([c * mech.eps_pureDP for (mech, c)
                                  in zip(mechanism_list, coeff_list)])
        newmech.delta0 = max([mech.delta0 for (mech, c)
                              in zip(mechanism_list, coeff_list)])

        if not RDP_compose_only:  # Also do KOV-composition while optimizing over \delta parameter
            # TODO: Also implement the KOV-composition here and propagate the updates
            # TODO: How do we generically compose eps(delta) functions?
            # TODO: How do we generically compose approximate RDP functions
            # TODO: How do we generically compose fDP? (efficiently)
            pass

        # Other book keeping
        newmech.name = self.update_name(mechanism_list, coeff_list)
        # keep track of all parameters of the composed mechanisms
        newmech.params = self.update_params(mechanism_list)

        return newmech

    def update_name(self,mechanism_list, coeff_list):
        separator = ', '
        s = separator.join([mech.name + ': ' + str(c) for (mech, c)
                           in zip(mechanism_list, coeff_list)])

        return 'Compose:{'+ s +'}'

    def update_params(self, mechanism_list):
        params = {}
        for mech in mechanism_list:
            params_cur = {mech.name+':'+k: v for k,v in mech.params.items()}
            params.update(params_cur)
        return params



# The generic composition class
class ComposeAFA(Transformer):
    """ The analytical Fourier Accountant (AFA) is a transformer that takes a list of Mechanisms and number of times they appear,
    and output a Mechanism that represents the composed mechanism.
    https://arxiv.org/pdf/2106.08567.pdf
    """
    def __init__(self):
        Transformer.__init__(self)
        self.name = 'ComposeFourier'

        # Update the function that is callable
        self.transform = self.compose

    def compose(self, mechanism_list, coeff_list):
        """
        In the composition, we keep track of two lists of characteristic functions (Phi(t) and Phi'(t))with
        respect to the privacy loss R.V. log(p/q) and log(q/p).
        For most basic mechanisms (e.g., Gaussian mechanism, Lapalce mechanism), their phi(t) and Phi'(t) are the same.
        For some advanced mechanisms (e.g., SubsampleGaussian mechanism), their characteristic functions are not symmetric.
        """

        newmech = Mechanism()

        # update the functions: log_phi_p and log_phi_q
        def new_log_phi_p2q(x):
            return sum([c * mech.log_phi_p2q(x) for (mech, c) in zip(mechanism_list, coeff_list)])

        def new_log_phi_q2p(x):
            return sum([c * mech.log_phi_q2p(x) for (mech, c) in zip(mechanism_list, coeff_list)])



        newmech.propagate_updates((new_log_phi_p2q, new_log_phi_q2p), 'log_phi')
        # Other book keeping
        newmech.name = self.update_name(mechanism_list, coeff_list)
        # keep track of all parameters of the composed mechanisms
        newmech.params = self.update_params(mechanism_list)

        return newmech

    def update_name(self,mechanism_list, coeff_list):
        separator = ', '
        s = separator.join([mech.name + ': ' + str(c) for (mech, c)
                           in zip(mechanism_list, coeff_list)])

        return 'Compose:{'+ s +'}'

    def update_params(self, mechanism_list):
        params = {}
        for mech in mechanism_list:
            params_cur = {mech.name+':'+k: v for k,v in mech.params.items()}
            params.update(params_cur)
        return params




# composition of only Gaussian mechanisms
class ComposeGaussian(Composition):
    """ CompositionGaussian is a specialized composation function of ONLY Guassian mechanisms
    output a Mechanism that represents the composed mechanism"""
    def __init__(self):
        Composition.__init__(self)
        self.name = 'ComposeGaussian'

    def compose(self, mechanism_list, coeff_list, RDP_compose_only=True, BBGHS_conversion=True, fDP_based_conversion=False):
        # Make sure that the list contains only Gaussian mechanisms
        for mech in mechanism_list:
            assert(isinstance(mech, mechanism_zoo.GaussianMechanism)
                   or isinstance(mech, mechanism_zoo.ExactGaussianMechanism))

        # Directly compose the distribution of privacy-loss random variables.
        # Sum of Gaussians are gaussians
        tmp = 0
        for mech, coeff in zip(mechanism_list,coeff_list):
            tmp += mech.params['sigma']**(-2)*coeff

        newmech = mechanism_zoo.ExactGaussianMechanism(sigma=math.sqrt(1/tmp))

        # Other book keeping
        newmech.name = self.update_name(mechanism_list, coeff_list)

        # keep track of all parameters of the composed mechanisms
        newmech.params = self.update_params(mechanism_list)

        return newmech





class AmplificationBySampling_pld(Transformer):
    def __init__(self, PoissonSampling=True, neighboring='add_only'):
        # Amplification By Sampling rule for PLD formalism
        # By default, poisson sampling is used:  sample a dataset by selecting each data point iid
        # If PoissonSampling is set to False, then it chooses a random subset with size prob * n
        # Compared to other amplification rules (e.g., rules in RDP), this rule treats neighboring relationship separately:
        # this amplification rule is for either "add_only" or "remove_only" relationship.
        Transformer.__init__(self)
        if PoissonSampling:
            self.name = 'PoissonSample'
        else:
            self.name = 'Subsample'
        self.PoissonSampling = PoissonSampling
        self.unary_operator = True
        self.preprocessing = True # Sampling happen before the mechanism is applied
        self.neighboring = neighboring

        # Update the function that is callable
        self.transform = self.amplify

    def amplify(self, mechanism, prob):
        # If the neighboring relationship is adding, the dominating distribution will be (P, (1-gamma)P + gamma*Q)
        # If the neighboring relationship is removing, the dominating distribution will be ( (1-gamma)*P + gamma*Q, P)

        newmech = Mechanism()

        # privacy amplification via approx-dp

        # Amplification of RDP
        # propagate to approxDP as well.
        # amplification by sampling for adding relationship is compatible with the base mechanism that has adding or
        # add/remove or replace relationship.
        if self.neighboring == 'add_only':
            assert  mechanism.neighboring != 'remove_only', "mechanism's neighboring relationship notion of DP is " \
                                                   "incompatible with Privacy Amplification by  sampling"
            # take care of the add neighbor and remove neighbor
            # If not, there actually isn't a way to convert it from replace-one notation,
            # unless a "dummy" user exists in the space.
            newmech.neighboring = 'add_only'
            # new_pdf_p and new_pdf_q denote the new dominating pair of sampled mechanism.
            new_pdf_p = lambda x: mechanism.pdf_p(x)
            new_pdf_q = lambda x: (1. - prob) * mechanism.pdf_p(x) + prob * mechanism.pdf_q(x)
            newmech.propagate_updates((new_pdf_p, new_pdf_q),'pdf')
            # Besides using general pdf2phi conversion, we can construct a more numerical stable phi function.
            # In the subsampled Gaussian case, the following phi functions is more stable as many exponential terms will
            # cancel out in the phi function.
            # Check if the base mechanism is Gaussian mechanism. If so, use available more-stable phi functions.
            if len(mechanism.params.keys()) == 1 and 'sigma' in mechanism.params.keys():
                params = {'sigma': mechanism.params['sigma'], 'gamma':prob}
                log_phi_p = lambda x: phi_bank.phi_subsample_gaussian_p(params, x, remove_only=False)
                log_phi_q = lambda x: phi_bank.phi_subsample_gaussian_q(params, x, remove_only = False)
                newmech.propagate_updates((log_phi_p, log_phi_q), 'log_phi')

        else:
            assert mechanism.neighboring != 'add_only', "mechanism's neighboring relationship notion of DP is " \
                                                   "incompatible with Privacy Amplification "
            newmech.neighboring ='remove_only'
            if self.PoissonSampling:
                new_pdf_p = lambda x: (1. - prob) * mechanism.pdf_q(x) + prob * mechanism.pdf_p(x)
                new_pdf_q = lambda x: mechanism.pdf_q(x)
                newmech.propagate_updates((new_pdf_p, new_pdf_q), 'pdf')
                # Check if the base mechanism is Gaussian mechanism. If so, use available more-stable phi functions.
                if len(mechanism.params.keys()) == 1 and 'sigma'in mechanism.params.keys():
                    params = {'sigma': mechanism.params['sigma'], 'gamma': prob}
                    log_phi_p2q = lambda x: phi_bank.phi_subsample_gaussian_p(params, x, remove_only=True)
                    log_phi_q2p = lambda x: phi_bank.phi_subsample_gaussian_q(params, x, remove_only=True)
                    newmech.propagate_updates((log_phi_p2q, log_phi_q2p), 'log_phi')
            else:
                # subset sampling
                new_pdf_p = lambda x: (1. - prob) * mechanism.pdf_p(x) + prob * mechanism.pdf_q(x)
                new_pdf_q = lambda x: mechanism.pdf_p(x)
        # convert pdf to phi
        newmech.propagate_updates([new_pdf_p, new_pdf_q], 'pdf')

        if prob == 0:
            new_approxDP = lambda delta:0
        else:
            new_approxDP = lambda delta: np.log(1 + prob*(np.exp(mechanism.approxDP(delta/prob))-1))
        newmech.approxDP = new_approxDP


        # TODO: For future work, we need to update subsampled CDF


        #TODO: Implement the amplification of f-DP
        # propagate to approxDP, or simply get the f-DP from approximate-DP.


        # book keeping
        key = self.name + '_' + str(prob)
        num = 0
        newname = self.name

        # the following handles the case when key is already in the params
        while key in mechanism.params:
            num = num+1
            newname = self.name+str(num)
            key = newname + '_' + str(prob)

        newmech.name = newname +':'+mechanism.name
        newmech.params = deepcopy(mechanism.params)
        new_params = {newname:prob}
        newmech.params.update(new_params)

        return newmech



class AmplificationBySampling(Transformer):
    def __init__(self, PoissonSampling=True):
        # By default, poisson sampling is used:  sample a dataset by selecting each data point iid
        # If PoissonSampling is set to False, then it chooses a random subset with size prob * n
        Transformer.__init__(self)
        if PoissonSampling:
            self.name = 'PoissonSample'
        else:
            self.name = 'Subsample'
        self.PoissonSampling = PoissonSampling
        self.unary_operator = True
        self.preprocessing = True # Sampling happen before the mechanism is applied

        # Update the function that is callable
        self.transform = self.amplify

    def amplify(self, mechanism, prob, improved_bound_flag=False):
        # If you know that your mechanism
        #  - (for PoissonSampling) satisfies the the conditions in Theorem 8 of http://proceedings.mlr.press/v97/zhu19c/zhu19c.pdf
        #  - or (for subsampling)  satisfies the conditions of Theorem 27 of https://arxiv.org/pdf/1808.00087.pdf
        # then you may switch general_bound_flag to False to get a tighter bound.

        # Else, for all mechanisms with RDP bounds, the general upper bounds are used by default.

        newmech = Mechanism()

        # privacy amplification via approx-dp



        # Amplification of RDP
        # propagate to approxDP as well.

        if self.PoissonSampling:
            assert mechanism.neighboring != 'replace_one', "mechanism's replace_one notion of DP is " \
                                                   "incompatible with Privacy Amplification " \
                                                   "by Poisson sampling"
            # check that the input mechanism uses the standard add-or-remove notion of DP.
            # If not, there actually isn't a way to convert it from replace-one notation,
            # unless a "dummy" user exists in the space.
            newmech.replace_one = False

        else:  # if we want subset subsampled DP
            assert mechanism.neighboring != 'add_remove', "mechanism's add-remove notion of DP is " \
                                                   "incompatible with Privacy Amplification " \
                                                   "by subsampling without replacements"
            # TODO: implement a transformer that convert add/remove to replace_one notion of DP.
            newmech.replace_one = True

        if prob == 0:
            new_approxDP = lambda delta:0
        else:
            new_approxDP = lambda delta: np.log(1 + prob*(np.exp(mechanism.approxDP(delta/prob))-1))
        newmech.approxDP = new_approxDP

        acct = rdp_acct.anaRDPacct()
        if self.PoissonSampling:
            if improved_bound_flag:
                acct.compose_poisson_subsampled_mechanisms(mechanism.RenyiDP,prob)
            else:
                acct.compose_poisson_subsampled_mechanisms1(mechanism.RenyiDP,prob)
        else:  # subsampling
            if improved_bound_flag:
                acct.compose_subsampled_mechanism(mechanism.RenyiDP, prob, improved_bound_flag=True)
            else:
                acct.compose_subsampled_mechanism(mechanism.RenyiDP, prob)


        acct.build_zeroth_oracle()
        new_rdp = acct.evalRDP
        newmech.propagate_updates(new_rdp,'RDP')

        #TODO: Implement the amplification of f-DP
        # propagate to approxDP, or simply get the f-DP from approximate-DP.


        # book keeping
        key = self.name + '_' + str(prob)
        num = 0
        newname = self.name

        # the following handles the case when key is already in the params
        while key in mechanism.params:
            num = num+1
            newname = self.name+str(num)
            key = newname + '_' + str(prob)

        newmech.name = newname +':'+mechanism.name
        newmech.params = deepcopy(mechanism.params)
        new_params = {newname:prob}
        newmech.params.update(new_params)

        return newmech




# TODO: implement other transformers:
# - amplification by shuffling
# - parallel composition
# - group composition
# - private selection of private candidates
# - amplification by overwhelmingly large-probability event.


"""
'mechanism_zoo' module implements popular DP mechanisms with their privacy guarantees

"""
import math
from scipy import special
import numpy as np
from autodp.autodp_core import Mechanism
from autodp import rdp_bank, dp_bank, fdp_bank, phi_bank
from autodp import transformer_zoo
from scipy.stats import norm
from scipy.optimize import minimize_scalar
import numpy as np


# Example of a specific mechanism that inherits the Mechanism class
class GaussianMechanism(Mechanism):
    """
    The example of Gaussian mechanism with different characterizations.
    """
    def __init__(self, sigma, name='Gaussian',
                 RDP_off=False, approxDP_off=False, fdp_off=True,
                 use_basic_RDP_to_approxDP_conversion=False,
                 use_fDP_based_RDP_to_approxDP_conversion=False, phi_off=True):
        """
        sigma: the std of the noise divide by the l2 sensitivity.
        coeff: the number of composition
        RDP_off: if False, then we characterize the mechanism using RDP.
        fdp_off: if False, then we characterize the mechanism using fdp.
        phi_off: if False, then we characterize the mechanism using phi-function.
        """
        Mechanism.__init__(self)

        self.name = name # When composing
        self.params = {'sigma': sigma} # This will be useful for the Calibrator
        # TODO: should a generic unspecified mechanism have a name and a param dictionary?

        self.delta0 = 0

        if not phi_off:
            """
            Apply phi function to analyze Gaussian mechanism.
            the CDF of privacy loss R.V. is computed using an integration (see details in cdf_bank) through Levy Theorem.
            """
            log_phi = lambda x: phi_bank.phi_gaussian({'sigma': sigma}, x)
            self.exact_phi = True
            # self.cdf tracks the cdf of log(p/q) and the cdf of log(q/p).
            self.propagate_updates((log_phi, log_phi), 'log_phi')

            # Propagate the pdf of dominating pairs.
            def pdf_p(x): return norm.pdf((x-1.), scale=sigma**2)
            def pdf_q(x): return norm.pdf(x, scale=sigma**2)
            self.propagate_updates((pdf_p, pdf_q), 'pdf')
            """
            Moreover, we know the closed-form expression of the CDF of the privacy loss RV
               privacy loss RV distribution l=log(p/q) ~ N(1/2\sigma^2, 1/sigma^2)
            We can also use the following closed-form cdf directly.
            """
            #sigma = sigma*1.0/np.sqrt(coeff)
            #mean = 1.0 / (2.0 * sigma ** 2)
            #std = 1.0 / (sigma)
            #cdf = lambda x: norm.cdf((x - mean) / std)
            #self.propagate_updates(cdf, 'cdf', take_log=True)


        if not RDP_off:
            new_rdp = lambda x: rdp_bank.RDP_gaussian({'sigma': sigma}, x)
            if use_fDP_based_RDP_to_approxDP_conversion:
                # This setting is slightly more complex, which involves converting RDP to fDP,
                # then to eps-delta-DP via the duality
                self.propagate_updates(new_rdp, 'RDP', fDP_based_conversion=True)
            elif use_basic_RDP_to_approxDP_conversion:
                self.propagate_updates(new_rdp, 'RDP', BBGHS_conversion=False)
            else:
                # This is the default setting with fast computation of RDP to approx-DP
                self.propagate_updates(new_rdp, 'RDP')

        if not approxDP_off: # Direct implementation of approxDP
            new_approxdp = lambda x: dp_bank.get_eps_ana_gaussian(sigma, x)
            self.propagate_updates(new_approxdp,'approxDP_func')

        if not fdp_off: # Direct implementation of fDP
            fun1 = lambda x: fdp_bank.log_one_minus_fdp_gaussian({'sigma': sigma}, x)
            fun2 = lambda x: fdp_bank.log_neg_fdp_grad_gaussian({'sigma': sigma}, x)
            self.propagate_updates([fun1,fun2],'fDP_and_grad_log')
            # overwrite the fdp computation with the direct computation
            self.fdp = lambda x: fdp_bank.fDP_gaussian({'sigma': sigma}, x)

        # the fDP of gaussian mechanism is equivalent to analytical calibration of approxdp,
        # so it should have been automatically handled numerically above


        # Discussion:  Sometimes delta as a function of eps has a closed-form solution
        # while eps as a function of delta does not
        # Shall we represent delta as a function of eps instead?


class ExactGaussianMechanism(Mechanism):
    """
    The Gaussian mechanism to use in practice with tight direct computation of everything
    """
    def __init__(self, sigma=None, name='Gaussian'):
        # the sigma parameter is the std of the noise divide by the l2 sensitivity
        Mechanism.__init__(self)

        self.name = name # When composing
        self.params = {'sigma': sigma} # This will be useful for the Calibrator
        self.delta0 = 0
        if sigma is not None:
            new_rdp = lambda x: rdp_bank.RDP_gaussian({'sigma': sigma}, x)
            self.propagate_updates(new_rdp, 'RDP')
            # Overwrite the approxDP and fDP with their direct computation
            self.approxDP = lambda x: dp_bank.get_eps_ana_gaussian(sigma, x)
            self.fDP = lambda x: fdp_bank.fDP_gaussian({'sigma': sigma}, x)


class LaplaceMechanism(Mechanism):
    """
    The Laplace Mechanism that support RDP and phi-function based characterization.
    """
    def __init__(self, b=None, name='Laplace', RDP_off=False, phi_off=True):
        """
        b: the ratio of the scale parameter and L1 sensitivity.
        RDP_off: if False, then we characterize the mechanism using RDP.
        fdp_off: if False, then we characterize the mechanism using fdp.
        phi_off: if False, then we characterize the mechanism using phi-function.
        """
        Mechanism.__init__(self)

        self.name = name
        self.params = {'b': b} # This will be useful for the Calibrator

        self.delta0 = 0
        if not phi_off:

            log_phi_p2q = lambda x: phi_bank.phi_laplace(self.params, x)
            log_phi_q2p = lambda x: phi_bank.phi_laplace(self.params, x)

            self.log_phi_p2q = log_phi_p2q
            self.log_phi_q2p = log_phi_q2p
            self.propagate_updates((log_phi_p2q, log_phi_q2p), 'log_phi')


        if not RDP_off:
            new_rdp = lambda x: rdp_bank.RDP_laplace({'b': b}, x)
            self.propagate_updates(new_rdp, 'RDP')



class RandresponseMechanism(Mechanism):

    """
    The randomized response mechanism that supports RDP and phi-function based characterization.
    TODO: assert when p is None.
    """
    def __init__(self, p=None, RDP_off=False, phi_off=True, name='Randresponse'):
        """
        p: the Bernoulli probability p of outputting the truth.
        """
        Mechanism.__init__(self)

        self.name = name
        self.params = {'p': p}
        self.delta0 = 0

        if not RDP_off:
            new_rdp = lambda x: rdp_bank.RDP_randresponse({'p': p}, x)
            self.propagate_updates(new_rdp, 'RDP')
            approxDP = lambda x: dp_bank.get_eps_randresp_optimal(p, x)
            self.propagate_updates(approxDP, 'approxDP_func')

        if not phi_off:
            log_phi = lambda x: phi_bank.phi_rr_p({'p': p, 'q':1-p}, x)
            self.propagate_updates((log_phi, log_phi), 'log_phi')

class zCDP_Mechanism(Mechanism):
    def __init__(self,rho,xi=0,name='zCDP_mech'):
        Mechanism.__init__(self)

        self.name = name
        self.params = {'rho':rho,'xi':xi}
        new_rdp = lambda x: rdp_bank.RDP_zCDP(self.params, x)

        self.propagate_updates(new_rdp,'RDP')


class ExponentialMechanism(zCDP_Mechanism):
    def __init__(self, eps, RDP_off=False, phi_off=True, non_adaptive=False, name='ExpMech'):
        zCDP_Mechanism.__init__(self, eps**2/8, name=name)
        # the zCDP bound is from here: https://arxiv.org/pdf/2004.07223.pdf

        self.eps_pureDP = eps
        self.propagate_updates(eps, 'pureDP')

        if not RDP_off:
            new_rdp = lambda x: rdp_bank.RDP_pureDP({'eps': eps}, x)
            self.propagate_updates(new_rdp, 'RDP')

        def func(t):
            """
            Return the bernoulli parameter p and q for any t.

            """
            p = (np.exp(-t) - np.exp(-eps))/(1-np.exp(-eps))
            q = (1 - np.exp(t-eps))/(1-np.exp(-eps))
            return {'p':p, 'q':q}
        # TODO: implement the f-function and phi-function representation from two logistic r.v.


        if not phi_off:
            # t is from [0, eps], log(p/q) in [t-eps, t], we optimize t over compositions.
            #  see page 10 for the dominating pair distribution. https://arxiv.org/pdf/1909.13830.pdf
            log_phi_p2q = lambda x, t: phi_bank.phi_rr_p(func(t), x)
            log_phi_q2p = lambda x, t: phi_bank.phi_rr_q(func(t), x)
            self.tbd_range = [0, eps]
            self.propagate_updates((log_phi_p2q, log_phi_q2p), 'log_phi_adv')




class PureDP_Mechanism(Mechanism):
    def __init__(self, eps, name='PureDP'):
        # the eps parameter is the pure DP parameter of this mechanism
        Mechanism.__init__(self)

        self.name = name # Used for generating new names when composing
        self.params = {'eps': eps} #

        self.propagate_updates(eps, 'pureDP')

        # ------- I verified that the following options give the same results ----
        # def new_rdp(x):
        #     return rdp_bank.RDP_pureDP({'eps': eps}, x)
        #
        # if use_basic_RDP_to_approxDP_conversion:
        #     self.propagate_updates(new_rdp, 'RDP', BBGHS_conversion=False)
        # else:


        #     self.propagate_updates(new_rdp, 'RDP')

class zCDP_Mechanism(Mechanism):
    def __init__(self,rho,xi=0,name='zCDP_mech'):
        Mechanism.__init__(self)

        self.name = name
        self.params = {'rho':rho,'xi':xi}
        new_rdp = lambda x: rdp_bank.RDP_zCDP(self.params, x)

        self.propagate_updates(new_rdp,'RDP')


class DiscreteGaussianMechanism(zCDP_Mechanism):
    def __init__(self, sigma, name='DGM'):
        zCDP_Mechanism.__init__(self, 0.5/sigma**2, name=name)

        # This the the best implementation for DGM for now.
        # The analytical formula for approximate-DP applies only to 1D outputs
        # The exact eps,delta-DP and char function for the multivariate outputs
        # requires a further maximization over neighboring datasets, which is unclear how to do

class ExponentialMechanism(zCDP_Mechanism):
    def __init__(self, eps, name='ExpMech'):
        zCDP_Mechanism.__init__(self, eps**2/8, name=name)
        # the zCDP bound is from here: https://arxiv.org/pdf/2004.07223.pdf
        self.params['eps'] = eps

        # TODO: Bounded range should imply a slightly stronger RDP that dominates the following
        self.eps_pureDP = eps
        self.propagate_updates(eps, 'pureDP')

        # TODO: implement the f-function and phi-function representation from two logistic r.v.



class SubsampleGaussianMechanism(Mechanism):
    """
    Poisson subsampled Gaussian mechanism.

    unlike the general mechanism design, we specify the number of composition as an input to initialize the mechanism.
    Therefore, we can use this mechanism as an example for privacy calibration: calibrate the noise scale for a subsample
    gaussian mechanism that runs for 'coeff' rounds.

    """
    def __init__(self,params, phi_off=True, RDP_off=False, neighboring='remove_only', name='SubsampleGaussian'):
        Mechanism.__init__(self)
        self.name=name
        self.params={'prob':params['prob'],'sigma':params['sigma'],'coeff':params['coeff']}

        if not RDP_off:
            # create such a mechanism as in previously
            subsample = transformer_zoo.AmplificationBySampling()  # by default this is using poisson sampling
            mech = GaussianMechanism(sigma=params['sigma'])

            # Create subsampled Gaussian mechanism
            SubsampledGaussian_mech = subsample(mech, params['prob'], improved_bound_flag=True)

            # Now run this for niter iterations
            compose = transformer_zoo.Composition()
            mech = compose([SubsampledGaussian_mech], [params['coeff']])

            # Now we get it and let's extract the RDP function and assign it to the current mech being constructed
            rdp_total = mech.RenyiDP
            self.propagate_updates(rdp_total, type_of_update='RDP')

        if not phi_off:
            """
            # phi-function based characterization takes care of add/remove neighboring relationship separately, see
            # https://arxiv.org/pdf/2106.08567.pdf
            # We can obtain the results for the standard add/remove by a pointwise maximum of the two.
            # For example, we define mech_add and mech_remove as follow:
            mech_add = SubsampleGaussianMechanism({'prob':prob, 'sigma':sigma, 'coeff': coeff}, phi_off=False,
             neighboring='add_only')
            mech_remove = SubsampleGaussianMechanism({'prob':prob, 'sigma':sigma, 'coeff': coeff}, phi_off=False,
             neighboring='remove_only')
            # Query epsilon at delta for the standard add/remove:
            eps = max(mech_add.approxDP(delta), mch_remove.approxDP(delta))
            
            """
            n_comp = params['coeff']
            if neighboring == 'remove_only':
                # log_phi_p2q denotes the approximated phi-function of the privacy loss R.V. log(p/q).
                # log_phi_q2p denotes the approximated phi-function of the privacy loss R.V. log(q/p).
                composed_log_phi_p2q = lambda x: n_comp * phi_bank.phi_subsample_gaussian_p\
                    (self.params, x, remove_only=True)
                composed_log_phi_q2p = lambda x: n_comp * phi_bank.phi_subsample_gaussian_q\
                    (self.params, x, remove_only=True)
                self.neighboring = 'remove_only'

            else:
                # neighboring for add_only
                composed_log_phi_p2q = lambda x: n_comp * phi_bank.phi_subsample_gaussian_p\
                    (self.params, x, remove_only=False)
                composed_log_phi_q2p = lambda x: n_comp * phi_bank.phi_subsample_gaussian_q\
                    (self.params, x, remove_only=False)
                self.neighboring = 'add_only'
            self.propagate_updates((composed_log_phi_p2q, composed_log_phi_q2p), 'log_phi')


class ComposedGaussianMechanism(Mechanism):
    """
    This one is used as an example for calibrator with composed Gaussian mechanism
    """
    def __init__(self,params,name='SubsampleGaussian'):
        Mechanism.__init__(self)
        self.name=name
        self.params={'sigma':params['sigma'],'coeff':params['coeff']}
        # create such a mechanism as in previously

        mech = GaussianMechanism(sigma=params['sigma'])
        # Now run this for coeff iterations
        compose = transformer_zoo.Composition()
        mech = compose([mech], [params['coeff']])

        # Now we get it and let's extract the RDP function and assign it to the current mech being constructed
        rdp_total = mech.RenyiDP
        self.propagate_updates(rdp_total, type_of_update='RDP')



class NoisyScreenMechanism(Mechanism):
    """
    The data-dependent RDP of ``Noisy Screening" (Theorem 7 in Private-kNN (CVPR-20))
    This mechanism is also used in Figure 2(a) in Gaussian-based SVTs.
    https://papers.nips.cc/paper/2020/file/e9bf14a419d77534105016f5ec122d62-Paper.pdf
    """
    def __init__(self,params,name='NoisyScreen'):
        Mechanism.__init__(self)
        self.name=name
        self.params={'logp':params['logp'],'logq':params['logq']}
        # create such a mechanism as in previously

        new_rdp = lambda x: rdp_bank.RDP_noisy_screen({'logp': params['logp'], 'logq': params['logq']}, x)
        self.propagate_updates(new_rdp, 'RDP')


class GaussianSVT_Mechanism(Mechanism):
    """
    The Gaussian-based SVT mechanism is described in
    https://papers.nips.cc/paper/2020/file/e9bf14a419d77534105016f5ec122d62-Paper.pdf

    The mechanism takes the parameter k and c. k is the maximum length before
    the algorithm stops. c is the cut-off parameter.  Setting rdp_c_1 = True implies that
    we use RDP-based Gaussian-SVT with c=1, else c>1.

    """
    def __init__(self,params,name='GaussianSVT', rdp_c_1=True):
        Mechanism.__init__(self)
        self.name=name
        if rdp_c_1 == True:
            self.name = name + 'c_1'
            valid_keys = ['sigma', 'k', 'margin', 'sigma_nu', 'Delta', 'sigma2', 'gamma']
            self.params = dict(filter(lambda tuple: tuple[0] in valid_keys, params.items()))
            new_rdp = lambda x: rdp_bank.RDP_gaussian_svt_c1(self.params, x)
        else:
            self.name = name + 'c>1'
            self.params = {'sigma':params['sigma'],'k':params['k'], 'c':params['c']}
            new_rdp = lambda x: rdp_bank.RDP_gaussian_svt_cgreater1(self.params, x)
        self.propagate_updates(new_rdp, 'RDP')

class LaplaceSVT_Mechanism(Mechanism):
    """
    Laplace SVT (c>=1) with a RDP description.
    b: the noise scale to perturb the threshold and the query.
    k: the maximum length before the algorithm steps. k could be infinite.

    We provide the RDP implementation and pure-DP implementation
    """
    def __init__(self,params,name='GaussianSVT'):
        Mechanism.__init__(self)
        self.name=name
        valid_keys = ['b', 'k', 'c']
        self.params = dict(filter(lambda tuple: tuple[0] in valid_keys, params.items()))
        new_rdp = lambda x: rdp_bank.RDP_svt_laplace(self.params, x)
        self.propagate_updates(new_rdp, 'RDP')


class StageWiseMechanism(Mechanism):
    """
    The StageWise generalized SVT is proposed by Zhu et.al., NeurIPS-20 (see Algorithm 3).
    This mechanism generalizes the Gaussian-based SVT.
    The mechanism takes two parameter c and k. c is the cut-off parameter (the number of tops in svt)
    and k denotes the  maximum limit for each chunk, e.g., the algorithm restarts whenever it encounters
     a top or reaches k limit.
    """
    def __init__(self, params=None,approxDP_off=False, name='StageWiseMechanism'):
        # the sigma parameter is the std of the noise divide by the l2 sensitivity
        Mechanism.__init__(self)

        self.name = name # When composing
        self.params = {'sigma': params['sigma'], 'k':params['k'], 'c':params['c']}
        self.delta0 = 0

        if not approxDP_off:  # Direct implementation of approxDP
            new_approxdp = lambda x: dp_bank.get_eps_gaussian_svt(params, x)
            self.propagate_updates(new_approxdp, 'approxDP_func')



# # Example 1: Short implementation of noisy gradient descent mechanism as a composition of GMs
# class NoisyGD_mech(GaussianMechanism):
#     def __init__(self,sigma_list,name='NoisyGD'):
#         GaussianMechanism.__init__(self, sigma=np.sqrt(np.sum(sigma_list)),name=name)
#         self.params = {'sigma_list':sigma_list}
#
# # The user could log sigma_list and then just declare a NoisyGD_mech object.
# mech = NoisyGD_mech(sigma_list)
# mech.get_approxDP(delta=1e-6)
#
#
# # Example 2: Implementing NoisySGD from basic building blocks
# subsample = Transformers.Subsample(prob=0.01)
# mech = Mechanisms.GaussianMechanism(sigma=5.0)
# # Create subsampled Gaussian mechanism
# SubsampledGaussian_mech = subsample(mech)
#
# # Now run this for 100 iterations
# compose = Transformers.Composition()
# NoisySGD_mech = compose(mechanism_list = [SubsampledGaussian_mech],coeffs_list=[100])
#
#
# # Example 3: You could also package this together by defining a NoisySGD mechanism
# class NoisySGD_mech(Mechanism):
#     def __init__(self,prob,sigma,niter,name='NoisySGD'):
#         Mechanism.__init__()
#         self.name=name
#         self.params={'prob':prob,'sigma':sigma,'niter':niter}
#
#         rdp = rdp_bank.subsampled_gaussian({'prob':params['prob'],'sigma':params['sigma']})
#         self.propagate_updates(rdp,type_of_update='RDP')
#
#
# # Example 4: Online decision. Hetereogenous sigma decided online
# # (maybe as a function of computed eps)
# # Alternatively if we want to do it via composition, so we can make online decision about
# # the sigma in the sigma_list
#
# delta = 1e-6
# online_sgd = Mechanisms.SubsampledGaussian_mech(prob=prob,sigma=sigma)
# compose = Transformers.Composition()
# for i in range(niter):
#     eps = online_ngd.get_approxDP(delta)
#     #determine the next prob, sigma
#     prob, sigma = func(eps)
#     mech_cur = Mechanisms.SubsampledGaussian_mech(prob=prob, sigma=sigma)
#     online_ngd = compose([online_ngd, mech_cur])
#
# # The above is quite general and can be viewed as a privacy accountant

class NoisySGD_Mechanism(Mechanism):
    def __init__(self, prob, sigma, niter, PoissonSampling=True, name='NoisySGD'):
        Mechanism.__init__(self)
        self.name = name
        self.params = {'prob': prob, 'sigma': sigma, 'niter': niter,
                       'PoissonSampling': PoissonSampling}

        # create such a mechanism as in previously
        subsample = transformer_zoo.AmplificationBySampling(PoissonSampling=PoissonSampling)
        # by default this is using poisson sampling

        mech = ExactGaussianMechanism(sigma=sigma)
        prob = prob
        # Create subsampled Gaussian mechanism
        SubsampledGaussian_mech = subsample(mech, prob, improved_bound_flag=True)
        # for Gaussian mechanism the improved bound always applies

        # Now run this for niter iterations
        compose = transformer_zoo.Composition()
        mech = compose([SubsampledGaussian_mech], [niter])

        # Now we get it and let's extract the RDP function and assign it to the current mech being constructed
        rdp_total = mech.RenyiDP
        self.propagate_updates(rdp_total, type_of_update='RDP')


class NoisyGD_Mechanism(GaussianMechanism):
    # With a predefined sequence of noise multipliers.
    def __init__(self,sigma_list, name='NoisyGD'):
        GaussianMechanism.__init__(self, sigma=np.sqrt(1/np.sum(1/sigma_list**2)), name=name)
        self.params = {'sigma_list': sigma_list}








"""
A collection of different ways to calibrate noise to privacy
"""
import numpy as np
from math import exp, sqrt
from scipy.special import erf
from scipy.optimize import brentq

from autodp import rdp_acct, rdp_bank


# Subsampling lemma and its inverse

def subsample_epsdelta(eps,delta,prob):
    """
    :param eps: privacy loss eps of the base mechanism
    :param delta: privacy loss delta of the base mechanism
    :param prob: subsampling probability
    :return: Amplified eps and delta

    This result applies to both subsampling with replacement and Poisson subsampling.

    The result for Poisson subsmapling is due to Theorem 1 of :
    Li, Ninghui, Qardaji, Wahbeh, and Su, Dong. On sampling, anonymization, and differential privacy or,
    k-anonymization meets differential privacy

    The result for Subsampling with replacement is due to:
    Jon Ullman's lecture notes: http://www.ccs.neu.edu/home/jullman/PrivacyS17/HW1sol.pdf
    See the proof of (b)

    """
    if prob == 0:
        return 0,0
    return np.log(1+prob*(np.exp(eps)-1)), prob*delta


def subsample_epsdelta_inverse(eps,delta,prob):
    # Give a target subsampled epsilon and delta, and subsampling probability. Calibrate the base eps, delta
    assert(prob > 0 and prob <=1)
    return np.log((np.exp(eps)-1)/prob + 1), np.minimum(delta/prob,1.0)


def subsample_epsdelta_get_prob(eps_target,delta_target, eps_base,delta_base):
    """
    Calibrate the probability of subsampling
    :param eps_target: Target eps in subsampled mechanisms
    :param delta_target: Target delta in subsampled mechanisms
    :param eps_base: base eps, in subsampled mechanisms
    :param delta_base: base delta in subsampled mechanisms.
    :return: subsampling probability  prob
    """
    return np.mininum(1.0,np.minimum(delta_target/delta_base,  (np.exp(eps_target)-1)/(np.exp(eps_base)-1)))



# we start with a general calibration function.

def RDP_mech(rdp_func, eps, delta, param_name, params, bounds=[0,np.inf],k=1,prob=1.0):
    # Take an analytical RDP, find the smallest noise level to achieve (eps, delta)-DP.
    """
    :param rdp_func: the RDP function that takes in params and alpha like those in 'rdp_bank'.
    :param eps:  the required eps
    :param delta:  the required delta
    :param param_name
    :param params: a template dictionary to modify from.
    :param bounds: a pair of numbers indicating the valid ranges of the parameters
    :return: params_out:  the calibrated params.
    """
    assert (eps > 0 and delta > 0)


    def func(x):
        # We assume that the rdp_func and param_name is chosen such that this function is either monotonically
        # increasing or decreasing.
        params[param_name] = x
        rdp = lambda alpha: rdp_func(params, alpha)
        tmp_acct = rdp_acct.anaRDPacct()

        if prob < 1.0 and prob >0:
            tmp_acct.compose_subsampled_mechanism(rdp, prob,coeff=k)
        else:
            tmp_acct.compose_mechanism(rdp,coeff=k)

        eps_tmp = tmp_acct.get_eps(delta)
        return eps_tmp - eps

    # Project to feasible region
    a=np.minimum(bounds[1], np.maximum(1.0, bounds[0]))
    b=np.maximum(bounds[0], np.minimum(2.0, bounds[1]))
    maxiter = 100
    count = 1
    # find a valid range
    while np.sign(func(a)) == np.sign(func(b)):
        a = np.maximum(a/2,bounds[0])
        b = np.minimum(b*2,bounds[1])
        count = count + 1
        if count >=maxiter:
            # infeasible
            raise ValueError('Infeasible privacy parameters for given RDP function and parameter bounds.')

    root = brentq(func, a, b)

    # assign calibarated results
    params_out = params
    params_out[param_name] = root

    return params_out

def subsampled_RDP_mech_get_prob(rdp_func, eps, delta, params,k=1):
    """
    This function calibrates the probability of subsampling to achieve a prescribed privacy goal.
    :param rdp_func: rdp of the base mechanism (as in those in rdp_bank)
    :param eps:  the required eps
    :param delta:  the required delta
    :param params: the parameter object for the first argument, rdp_func
    :param k: (optional) number of rounds to compose.
    :return:
    """
    assert (eps > 0 and delta > 0)
    def func(x):
        rdp = lambda alpha: rdp_func(params, alpha)
        tmp_acct = rdp_acct.anaRDPacct()
        tmp_acct.compose_subsampled_mechanism(rdp, x, coeff=k)
        eps_tmp = tmp_acct.get_eps(delta)
        return eps_tmp - eps

    root = brentq(func, 0, 1)
    return root


def gaussian_mech(eps,delta,k=1,prob=1.0):
    """
    Calibrate the scale parameter b of the Gaussian mechanism using RDP

    :param eps: prescribed eps
    :param delta: prescribed delta
    :param k: (optional) number of times to run this mechanism.
    :return: the parameter structure for this randomized algorithm
    """
    assert(eps>0)

    params = {}
    param_name = 'sigma'
    params = RDP_mech(rdp_bank.RDP_gaussian, eps, delta, param_name, params,k=k,prob=prob)

    return params



def laplace_mech(eps, delta, k=1, prob=1.0):
    """
    Calibrate the scale parameter b of the Laplace mechanism

    :param eps: prescribed eps
    :param delta: prescribed delta
    :param k: (optional) number of times to run this mechanism.
    :return: the parameter structure for this randomized algorithm
    """
    assert(eps>0 and delta >= 0)

    binf = 1.0/eps*k
    if delta == 0:
        params = {'b':binf}
    else:
        params = {}
        param_name = 'b'
        params = RDP_mech(rdp_bank.RDP_laplace, eps, delta, param_name, params, k=k,prob=prob)
        if params['b'] > binf:
            # further take maximum w.r.t. alpha = infty
            params['b'] = binf

    return params


def randresponse_mech(eps, delta, k=1,prob=1.0):
    """
    Calibrate the bernoulli parameter p of the randomized response mechanisms

    :param eps: prescribed eps
    :param delta: prescribed delta
    :param k: (optional) number of times to run this mechanism.
    :return: the parameter structure for this randomized algorithm
    """
    assert (eps > 0 and delta >= 0)

    pinf= np.exp(1.0*eps/k)/(1+np.exp(1.0*eps/k))
    if delta ==1:
        return {'p':1}
    if delta == 0:
        params = {'p':pinf}
    else:
        params = {}
        param_name = 'p'
        params = RDP_mech(rdp_bank.RDP_randresponse, eps, delta, param_name, params,
                          bounds=[np.exp(1.0*eps/k/2)/(1+np.exp(1.0*eps/2/k)),1-1e-8],k=k,prob=prob)
        if params['p'] < pinf:
            # further take maximum w.r.t. alpha = infty
            params['p'] = pinf

    return params


def classical_gaussian_mech(eps,delta):
    """
    The classical gaussian mechanism. For benchmarking purposes only.
    DO NOT USE in practice as it is dominated by `ana_gaussian_mech' and `gaussian_mech`.

    :param eps: prescribed 0< eps <1
    :param delta: prescribed 0 < delta <1
    :return: required noise level.
    """
    assert(eps > 0 and eps <=1),\
        "The classical Gaussian mechanism only supports 0 < eps <1, try `gaussian_mech` and `ana_gaussian_mech`"
    if delta <= 0:
        return np.inf
    if delta >= 1:
        return 0
    params = {'sigma': 1.0 * np.sqrt(2 * np.log(1.25 / delta)) / eps}
    return params


def ana_gaussian_mech(epsilon, delta, tol=1.e-12):
    """ Calibrate a Gaussian perturbation for differential privacy using the analytic Gaussian mechanism of [Balle and Wang, ICML'18]

    Modified from https://github.com/BorjaBalle/analytic-gaussian-mechanism/blob/master/agm-example.py

    Arguments:
    epsilon : target epsilon (epsilon > 0)
    delta : target delta (0 < delta < 1)
    tol : error tolerance for binary search (tol > 0)
    Output:
    params : a dictionary that contains field `sigma' --- the standard deviation of Gaussian noise needed to achieve
        (epsilon,delta)-DP under global sensitivity 1
    """

    def Phi(t):
        return 0.5*(1.0 + erf(float(t)/sqrt(2.0)))

    def caseA(epsilon,s):
        return Phi(sqrt(epsilon*s)) - exp(epsilon)*Phi(-sqrt(epsilon*(s+2.0)))

    def caseB(epsilon,s):
        return Phi(-sqrt(epsilon*s)) - exp(epsilon)*Phi(-sqrt(epsilon*(s+2.0)))

    def doubling_trick(predicate_stop, s_inf, s_sup):
        while(not predicate_stop(s_sup)):
            s_inf = s_sup
            s_sup = 2.0*s_inf
        return s_inf, s_sup

    def binary_search(predicate_stop, predicate_left, s_inf, s_sup):
        s_mid = s_inf + (s_sup-s_inf)/2.0
        while(not predicate_stop(s_mid)):
            if (predicate_left(s_mid)):
                s_sup = s_mid
            else:
                s_inf = s_mid
            s_mid = s_inf + (s_sup-s_inf)/2.0
        return s_mid

    delta_thr = caseA(epsilon, 0.0)

    if (delta == delta_thr):
        alpha = 1.0
        sigma = alpha / sqrt(2.0 * epsilon)

    else:
        if (delta > delta_thr):
            predicate_stop_DT = lambda s : caseA(epsilon, s) >= delta
            function_s_to_delta = lambda s : caseA(epsilon, s)
            predicate_left_BS = lambda s : function_s_to_delta(s) > delta
            function_s_to_alpha = lambda s : sqrt(1.0 + s/2.0) - sqrt(s/2.0)

        else:
            predicate_stop_DT = lambda s : caseB(epsilon, s) <= delta
            function_s_to_delta = lambda s : caseB(epsilon, s)
            predicate_left_BS = lambda s : function_s_to_delta(s) < delta
            function_s_to_alpha = lambda s : sqrt(1.0 + s/2.0) + sqrt(s/2.0)

        predicate_stop_BS = lambda s : abs(function_s_to_delta(s) - delta) <= tol

        s_inf, s_sup = doubling_trick(predicate_stop_DT, 0.0, 1.0)
        s_final = binary_search(predicate_stop_BS, predicate_left_BS, s_inf, s_sup)
        alpha = function_s_to_alpha(s_final)
        sigma = alpha/sqrt(2.0*epsilon)


    params = {'sigma': sigma}
    return params





"""
This module implements all known conversions from DP
1. puredp_to_approxdp
2. rdp_to_delta and rdp_to_aapproxDP
3. For  fDP conversions, we implement:
    a. single_rdp_to_fdp
    b. approxdp_func_to_fdp
    c. puredp_to_fdp(eps) # From Wasserman and Zhou
4. For phi function to approxDP conversions, we implement:
    a. pdf_to_phi (convert pdf of privacy loss R.V. to the characteristic function of privacy loss R.V>)
    b. cdf_to_approxdp and cdf_to_approxdelta
    c. phi_to_cdf
"""


import numpy as np
from autodp import utils
import math
import scipy.integrate as integrate
from autodp import rdp_bank
from scipy.optimize import minimize_scalar, root_scalar
from autodp import utils
from scipy.fft import fft
import time

def puredp_to_rdp(eps):
    # From pure dp to RDP
    assert(eps >= 0)

    def rdp(alpha):
        assert(alpha >= 0)
        if alpha==1:
            # Calculate this by l'Hospital rule
            return eps*(math.cosh(eps)-1)/math.sinh(eps)
        elif np.isinf(alpha):
            return eps
        elif alpha>1:
            # in the proof of Lemma 4 of Bun et al. (2016)
            s, mag = utils.stable_log_diff_exp(utils.stable_log_sinh(alpha*eps),
                                               utils.stable_log_sinh((alpha-1)*eps))
            return (mag - utils.stable_log_sinh(eps))/(alpha-1)
        else:
            return min(alpha * eps * eps /2, eps*(math.cosh(eps)-1)/math.sinh(eps))

    return rdp

def puredp_to_fdp(eps):
    # From Wasserman and Zhou
    def fdp(fpr):
        return np.max(np.array([0, 1-np.exp(eps)*fpr, np.exp(-eps)*(1-fpr)]))
    return fdp


def puredp_to_approxdp(eps):
    # Convert pureDP to approx dp
    # Page 3 of https://eprint.iacr.org/2018/277.pdf
    def approxdp(delta):
        s,mag = utils.stable_log_diff_exp(eps, np.log(delta))
        return mag
    return approxdp


def rdp_to_delta(rdp, BBGHS_conversion=True):
    """
    from RDP to delta with a fixed epsilon
    """


    def approx_delta(eps, naive=False):
        """
        approximate delta as a function of epsilon
        """
        def get_delta(alpha):
            if BBGHS_conversion:
                # Also by Canonne et al., 2020
                return np.minimum(
                    (np.exp((alpha-1)*(rdp(alpha)-eps))/alpha)*((1-1/alpha)**(alpha-1)), 1.0)
            else:
                # Naive conversion from Mironov
                return np.minimum(np.exp((alpha-1)*(rdp(alpha)-eps)),1.0)

        results = minimize_scalar(get_delta, method = 'Brent', bracket=(1,2))
        if results.success:
            # print('delta', delta,'eps under rdp', results.fun)
            return results.fun
        else:
            print('not found')
            return 1.0

        # def get_eps(delta):
        #     if delta == 0:
        #         return rdp(np.inf)
        #     else:
        #         def fun(x):  # the input the RDP's \alpha
        #             if x <= 1:
        #                 return np.inf
        #             else:
        #
        #                 if naive:
        #                     return np.log(1 / delta) / (x - 1) + rdp(x)
        #                 bbghs = np.maximum(rdp(x) + np.log((x-1)/x)
        #                                   - (np.log(delta) + np.log(x))/(x-1), 0)
        #                 """
        #                 The following is for optimal conversion
        #                 1/(alpha -1 )log(e^{(alpha-1)*rdp -1}/(alpha*delta) +1 )
        #                 """
        #                 sign, term_1= utils.stable_log_diff_exp((x-1)*rdp(x),0)
        #                 result = utils.stable_logsumexp_two(term_1 - np.log(x)- np.log(delta),0)
        #                 return min(result*1.0/(x - 1), bbghs)
        #
        #         results = minimize_scalar(fun, method='Brent', bracket=(1, 2))#, bounds=[1, 100000])
        #         if results.success:
        #            # print('delta', delta,'eps under rdp', results.fun)
        #             return results.fun
        #         else:
        #             return np.inf
        #
        #
        # def err(delta):
        #     current_eps = get_eps(delta)
        #     #print('current delta', delta, 'eps', current_eps)
        #     return abs(eps - current_eps)
        #
        # results = minimize_scalar(err, method='bounded', bounds=[0, 0.1],options={'xatol':1e-14})
        # if results.success:
        #     #print('results', results.x)
        #     return results.x
        # else:
        #     print('not found')
        #     return 1


    return approx_delta

def rdp_to_approxdp(rdp, alpha_max=np.inf, BBGHS_conversion=True):
    # from RDP to approx DP
    # alpha_max is an optional input which sometimes helps avoid numerical issues
    # By default, we are using the RDP to approx-DP conversion due to BBGHS'19's Theorem 21
    # paper: https://arxiv.org/pdf/1905.09982.pdf
    # if you need to use the simpler RDP to approxDP conversion for some reason, turn the flag off

    def approxdp(delta):
        """
        approxdp outputs eps as a function of delta based on rdp calculations

        :param delta:
        :return: the \epsilon with a given delta
        """

        if delta < 0 or delta > 1:
            print("Error! delta is a probability and must be between 0 and 1")
        if delta == 0:
            return rdp(np.inf)
        else:
            def fun(x):  # the input the RDP's \alpha
                if x <= 1:
                    return np.inf
                else:
                    if BBGHS_conversion:
                        return np.maximum(rdp(x) + np.log((x-1)/x)
                                          - (np.log(delta) + np.log(x))/(x-1), 0)
                    else:
                        return np.log(1 / delta) / (x - 1) + rdp(x)

            results = minimize_scalar(fun, method='Brent', bracket=(1,2))#, bounds=[1, alpha_max])
            if results.success:
                return results.fun
            else:
                # There are cases when certain \delta is not feasible.
                # For example, let p and q be uniform the privacy R.V. is either 0 or \infty and unless all \infty
                # events are taken cared of by \delta, \epsilon cannot be < \infty
                return np.inf
    return approxdp


def single_rdp_to_fdp(alpha, rho):
    assert (alpha >= 0.5)

    def fdp(x):
        assert (0 <= x <= 1)

        if x == 0:
            return 1
        elif x == 1:
            return 0

        if alpha == 1:
            # in this case rho is the KL-divergence
            def fun(y):
                assert (0 <= y <= 1 - x)
                if y == 0:
                    if x == 1:
                        return 0
                    else:
                        return np.inf
                elif y == 1:
                    if x == 0:
                        return 0
                    else:
                        return np.inf
                diff1 = (x * (np.log(x) - np.log(1 - y))
                         + (1 - x) * (np.log(1 - x) - np.log(y)) - rho)
                diff2 = (y * (np.log(y) - np.log(1 - x))
                         + (1 - y) * (np.log(1 - y) - np.log(x)) - rho)

                return np.maximum(diff1, diff2)
        else:
            # This is the general case for Renyi Divergence with \alpha > 1 or \alpha <1
            # find y such that
            # log(x^alpha (1-y)^{1-alpha} + (1-x)^alpha y^{1-alpha}) =  rho(alpha-1))
            # and log(y^alpha (1-x)^{1-alpha}) + (1-y)^alpha x^{1-alpha} =  rho(alpha-1))
            def fun(y):
                if y == 0:
                    if x == 1:
                        return 0
                    else:
                        return np.inf
                elif y == 1:
                    if x == 0:
                        return 0
                    else:
                        return np.inf

                diff1 = (utils.stable_logsumexp_two(alpha * np.log(x) + (1 - alpha) * np.log(1 - y),
                                                    alpha * np.log(1 - x) + (1 - alpha) * np.log(y))
                         - rho * (alpha - 1))
                diff2 = (utils.stable_logsumexp_two(alpha * np.log(y) + (1 - alpha) * np.log(1 - x),
                                                    alpha * np.log(1 - y) + (1 - alpha) * np.log(x))
                         - rho * (alpha - 1))
                if alpha > 1:
                    return np.maximum(diff1, diff2)
                else: # alpha < 1
                    # Notice that the sign of the inequality is toggled
                    return np.minimum(diff1, diff2)


        def normal_equation(y):
            # for finding the root
            return abs(fun(y))

        # there are two roots, we care about the roots smaller than 1-x
        results = minimize_scalar(normal_equation, bounds=[0, 1-x], method='bounded',
                                  options={'xatol':1e-9*(1-x)})
        if results.success:
            return results.x
        else:
            return 0.0
    return fdp


def rdp_to_fdp(rdp, alpha_max=np.inf):
    # RDP function to FDP.
    # maximize the fdp over alpha
    def fdp(x):
        assert (0 <= x <= 1)
        if x == 0:
            return 1
        elif x == 1:
            return 0

        def fun(alpha):
            if alpha < 0.5:
                return 0
            else:
                single_fdp = single_rdp_to_fdp(alpha, rdp(alpha))
                return -single_fdp(x)

        # This will use brent to start with 1,2.
        results = minimize_scalar(fun, bracket=(0.5, 2))#, bounds=(0.5, alpha_max))
        if results.success:
            return -results.fun
        else:
            return 0.0
    return fdp


def single_rdp_to_fdp_and_fdp_grad_log(alpha, rho):
    # Return two functions
    # the first function outputs log(1-fdp(x)) as a function of logx
    # the second function outputs log(-partial fdp(x)) as a function of logx # given epsilon, search a proper delta
    # The format of the output of the second function is an interval.
    assert (alpha >= 0.5)

    def diff1_KL(logx,u):
        assert(logx < u < 0)
        return (np.exp(logx) * (logx - u)
                + (1 - np.exp(logx)) * (np.log(1 - np.exp(logx)) - np.log(1-np.exp(u))) - rho)

    def diff2_KL(logx,u):
        return ((1 - np.exp(u)) * (np.log(1 - np.exp(u)) - np.log(1 - np.exp(logx)))
                + np.exp(u) * (u - logx) - rho)

    def diff1_general(logx,u):
        return (utils.stable_logsumexp_two(alpha * logx + (1 - alpha) * u,
                                           alpha * np.log(1 - np.exp(logx))
                                           + (1 - alpha) * np.log(1 - np.exp(u)))
         - rho * (alpha - 1))

    def diff2_general(logx,u):
        return (utils.stable_logsumexp_two(alpha * np.log(1-np.exp(u))
                                           + (1 - alpha) * np.log(1 - np.exp(logx)),
                                           alpha * u + (1 - alpha) * logx) - rho * (alpha - 1))

    def grad1_KL(logx,u):
        mag1 = np.log(u - logx + np.log(1-np.exp(logx)) - np.log(1-np.exp(u)))
        s, mag2 = utils.stable_log_diff_exp(np.log(1-np.exp(logx))- np.log(1-np.exp(u)),logx - u)
        return mag1 - mag2
        # return (logx - u - np.log(1-np.exp(logx))
        #         + np.log(1-np.exp(u))) / ((1-np.exp(logx))/(1-np.exp(u))
        #                                   - np.exp(logx)/np.exp(u))

    def grad2_KL(logx,u):
        mag1 = np.log(u - logx + np.log(1-np.exp(logx)) - np.log(1-np.exp(u)))
        s, mag2 = utils.stable_log_diff_exp(u-logx, np.log(1-np.exp(u))- np.log(1-np.exp(logx)))
        return mag2 - mag1
        # return ((1-np.exp(u))/(1-np.exp(logx))
        #         - np.exp(u)/np.exp(logx)) / (u - logx
        #                                     - np.log(1-np.exp(u)) + np.log(1-np.exp(logx)))

    def grad1_general(logx,u):
        #return - grad1_general(np.log(1-np.exp(u)), np.log(1-np.exp(logx)))

        s, mag = utils.stable_log_diff_exp(alpha * (np.log(1 - np.exp(logx))
                                                    - np.log(1 - np.exp(u))), alpha * (logx-u))
        if alpha > 1:
            s, mag1 = utils.stable_log_diff_exp((alpha-1) * (np.log(1 - np.exp(logx))
                                                             - np.log(1 - np.exp(u))),
                                                (alpha-1) * (logx-u))
            return np.log(alpha)-np.log(alpha-1) + mag1 - mag
        else:
            s, mag1 = utils.stable_log_diff_exp((alpha-1) * (logx-u),
                                                (alpha-1) * (np.log(1 - np.exp(logx))
                                                             - np.log(1 - np.exp(u))))
            return np.log(alpha)-np.log(1-alpha) + mag1 - mag

    def grad2_general(logx,u):
        s, mag = utils.stable_log_diff_exp(alpha * (u - logx),
                                           alpha * (np.log(1 - np.exp(u))
                                                    - np.log(1 - np.exp(logx))))
        if alpha > 1:
            s, mag2 = utils.stable_log_diff_exp((alpha-1) * (u - logx),
                                                (alpha-1) * (np.log(1 - np.exp(u))
                                                - np.log(1 - np.exp(logx))))
            return (np.log(1-1.0/alpha)) + mag - mag2
        else:  # if alpha < 1
            s, mag2 = utils.stable_log_diff_exp((alpha-1) * (np.log(1 - np.exp(u))
                                                - np.log(1 - np.exp(logx))),
                                                (alpha - 1) * (u - logx))
            return np.log(1.0/alpha - 1) + mag - mag2

    def log_one_minus_fdp(logx):
        #assert (0 <= x <= 1)
        assert(logx <= 0)

        if logx == 0: # x==1, f(x) should be 0
            return 0
        elif np.isneginf(logx): # x = 0,  f(x) should be 1
            return -np.inf

        # Now define the non-linear equation ``fun''
        # such that the u such that fun(u) = 0 gives log(1-f(x))

        if alpha == 1:
            # in this case rho is the KL-divergence
            def fun(u):
                assert( u >= logx)
                # assert (0 <= y <= 1 - x)
                if u == 0: #y == 0:
                    if logx == 0: #x == 1:
                        return 0
                    else:
                        return np.inf
                elif np.isneginf(u): #y == 1:
                    if np.isneginf(logx): #x == 0:
                        return 0
                    else:
                        return np.inf

                diff1 = diff1_KL(logx,u)
                #diff1 = (x * (np.log(x) - np.log(1 - y))
                #         + (1 - x) * (np.log(1 - x) - np.log(y)) - rho)
                diff2 = diff2_KL(logx,u)

                #diff2 = (y * (np.log(y) - np.log(1 - x))
                #         + (1 - y) * (np.log(1 - y) - np.log(x)) - rho)

                return np.maximum(diff1, diff2)
        else:
            # This is the general case for Renyi Divergence with \alpha > 1 or \alpha <1
            # find y such that
            # log(x^alpha (1-y)^{1-alpha} + (1-x)^alpha y^{1-alpha}) =  rho(alpha-1))
            # and log(y^alpha (1-x)^{1-alpha}) + (1-y)^alpha x^{1-alpha} =  rho(alpha-1))
            def fun(u):
                assert( u >= logx)
                if u == 0: #y == 0:
                    if logx == 0: #x == 1:
                        return 0
                    else:
                        return np.inf
                elif np.isneginf(u): #y == 1:
                    if np.isneginf(logx): #x == 0:
                        return 0
                    else:
                        return np.inf

                # diff1 = (utils.stable_logsumexp_two(alpha * np.log(x) + (1 - alpha) * np.log(1 - y),
                #                                     alpha * np.log(1 - x) + (1 - alpha) * np.log(y))
                #          - rho * (alpha - 1))
                diff1 = diff1_general(logx,u)
                # diff2 = (utils.stable_logsumexp_two(alpha * np.log(y) + (1 - alpha) * np.log(1 - x),
                #                                     alpha * np.log(1 - y) + (1 - alpha) * np.log(x))
                #          - rho * (alpha - 1))
                diff2 = diff2_general(logx,u)
                if alpha > 1:
                    return np.maximum(diff1, diff2)
                else: # alpha < 1
                    # Notice that the sign of the inequality is toggled
                    return np.minimum(diff1, diff2)

        def normal_equation(u):
            # for finding the root
            return abs(fun(u))

        # there are two roots, we care about the roots smaller than 1-x
        results = minimize_scalar(normal_equation, bounds=[logx,0], method='bounded',
                                  options={'xatol':1e-6})
        if results.success:
            return results.x
        else:
            return 0.0



    def log_neg_partial_fdp(logx):
        assert(logx <= 0)

        if np.isneginf(logx): # x = 0,  the gradient is negative infinity unless alpha = +inf
            # but alpha = +inf won't be passed into here.
            return  np.inf, np.inf
        elif logx == 0: # x = 1
            return 0, 0

        u = log_one_minus_fdp(logx)
        # Find which leg is active, and output the log (- subgradient)
        tol = 1e-5

        grad_l = np.inf
        grad_h = 0

        if alpha == 1:
            err = min(abs(diff1_KL(logx, u)), abs(diff2_KL(logx,u)))
            if err > tol:
                print('no solution found!')

            if abs(diff1_KL(logx,u)) <= tol:
                grad_l = grad1_KL(logx,u)
                grad_h = grad_l
            if abs(diff2_KL(logx,u)) <= tol:
                grad = grad2_KL(logx, u)
                grad_l = min(grad,grad_l)
                grad_h = max(grad,grad_h)
        else:
            err = min(abs(diff1_general(logx, u)), abs(diff2_general(logx,u)))
            if err > tol:
                print('no solution found!')
            if abs(diff1_general(logx,u)) <= tol:
                grad_l = grad1_general(logx,u)
                grad_h = grad_l
            if abs(diff2_general(logx,u)) <= tol:
                grad = grad2_general(logx,u)
                grad_l = min(grad,grad_l)
                grad_h = max(grad,grad_h)
        return [grad_l,grad_h]


    # ------------ debugging --------------------------
    # def fdp(x):
    #     return 1- np.exp(log_one_minus_fdp(np.log(x)))
    #
    # fdp_ref = single_rdp_to_fdp(alpha, rho)
    #
    # import matplotlib.pyplot as plt
    #
    # fpr_list = np.linspace(0,1,100)
    # plt.figure(1)
    # fnr1 = [fdp(x) for x in fpr_list]
    # fnr2 = [fdp_ref(x) for x in fpr_list]
    # plt.plot(fpr_list, fnr1)
    # plt.plot(fpr_list, fnr2)
    #
    # x = 0.01
    # u = log_one_minus_fdp(np.log(x))
    # log_minus_grad_ref = grad2_general(np.log(x), u)
    #
    # log_minus_grad = log_neg_partial_fdp(np.log(x))
    #
    #
    # def grad2_general_new(x,y):
    #     return - alpha*((x/(1-y))**(alpha-1) - ((1-x)/y)**(alpha-1)) / (1-alpha) / ( - (x/(1-y))**alpha + ((1-x)/y)**alpha)
    #
    # def Fxy(x,y):
    #     return (x/(1-y))**alpha * (1-y) +  ((1-x)/y)**alpha * y - np.exp((alpha-1)*rho)
    #
    # y = 1-np.exp(u)
    #
    # grad_ref = grad2_general_new(x,y)
    #
    # grad = -np.exp(log_minus_grad)
    # grad = grad[0]
    #
    # def tangent_line(v):
    #     return y + grad * (v - x)
    #
    # plt.plot(fpr_list, tangent_line(fpr_list))
    #
    # plt.ylim([0,1])
    # plt.xlim([0,1])
    #
    # plt.show()


    return log_one_minus_fdp, log_neg_partial_fdp


def rdp_to_fdp_and_fdp_grad_log(rdp, alpha_max=np.inf):
    # Return the a function that outputs the minimum of
    # log(1-fdp_alpha(x)) and the corresponding log(-partial fdp_alpha(x)) at the optimal \alpha.

    # This, when plugged into the standard machinery, would allow a more direct conversion from RDP.

    def log_one_minus_fdp(logx):
        assert (logx <= 0)
        if np.isneginf(logx):# x == 0:
            return [-np.inf, np.inf] # y = 1,  log (1-y) = -np.inf,  alpha is inf (pure DP)
        elif logx == 0:
            return [0, np.inf]

        def fun(alpha):
            if alpha < 0.5:
                return np.inf
            else:
                log_one_minus_fdp_alpha, tmp = single_rdp_to_fdp_and_fdp_grad_log(alpha, rdp(alpha))
                return log_one_minus_fdp_alpha(logx)

        # This will use brent to start with 1,2.
        results = minimize_scalar(fun, bracket=(0.5, 2))#, bounds=(0.5, alpha_max))
        if results.success:
            return [results.fun, results.x]
        else:
            return [log_one_minus_fdp(results.x), results.x]

    def log_one_minus_fdp_only(logx):
        res = log_one_minus_fdp(logx)
        return res[0]

    def log_neg_partial_fdp(logx):
        assert (logx <=0)
        if np.isneginf(logx):# x == 0:
            tmp = rdp(np.inf)
            return [tmp, np.inf] # y = 1,  log (1-y) = -np.inf,  alpha is inf (pure DP)
        elif logx == 0:
            tmp = rdp(np.inf)
            return [-np.inf, -tmp]

        # The following implements the more generic case
        # when we need to find the alpha that is active

        res = log_one_minus_fdp(logx)
        best_alpha = res[1]
        tmp, log_neg_partial_fdp_alpha = single_rdp_to_fdp_and_fdp_grad_log(best_alpha,
                                                                            rdp(best_alpha))
        return log_neg_partial_fdp_alpha(logx)

    return log_one_minus_fdp_only, log_neg_partial_fdp


def approxdp_to_approxrdp(eps,delta):
    # from a single eps,delta calculation to an approxdp function
    def approxrdp(alpha, delta1):
        if delta1 >= delta:
            rdp = puredp_to_rdp(eps)
            return rdp(alpha)
        else:
            return np.infty
    return approxrdp


def approxdp_func_to_approxrdp(eps_func):
    # from an approximate_dp function to approxrdp function
    def approxrdp(alpha, delta):
        rdp = puredp_to_rdp(eps_func(delta))
        return rdp(alpha)

    return approxrdp



def approxdp_to_fdp(eps, delta):
    # from a single eps, delta approxdp to fdp
    assert(eps >= 0 and 0 <= delta <= 1)

    def fdp(fpr):
        assert(0 <= fpr <= 1)
        if fpr == 0: # deal with log(0) below
            return 1-delta
        elif np.isinf(eps):
            return 0
        else:
            return np.max(np.array([0, 1-delta-np.exp(eps)*fpr, np.exp(-eps)*(1-delta-fpr)]))
    return fdp


def approxdp_func_to_fdp(func, delta_func=False):
    """
    from an approxdp function to fdp
    :param func: epsilon as a function of delta by default.
    :param delta_func: if the flag is True, then 'func' is a delta as a function of epsilon.
    :return: fdp function
    """
    #
    # By default, logdelta_func is False, and func is eps as a function of delta
    # fpr = maximize_{delta} approxdp_to_fdp(eps(delta),delta)(fpr)
    # if delta_func is True, it means that 'func' is a delta as a function of eps, then
    # fpr = maximize_{delta} approxdp_to_fdp(eps,delta(eps))(fpr)
    if delta_func:
        def fdp(fpr):

            assert(0 <= fpr <= 1)
            if fpr == 1:
                return 0

            def fun(eps):
                fdp_eps = approxdp_to_fdp(eps, func(eps))
                fnr = fdp_eps(fpr)
                return -fnr

            results = minimize_scalar(fun, bounds=[0, +np.inf], options={'disp': False})
            if results.success:
                return -results.fun
            else:
                return 0
    else:
        def fdp(fpr):
            assert(0 <= fpr <= 1)
            if fpr == 1:
                return 0

            def fun(delta):
                fdp_delta = approxdp_to_fdp(func(delta), delta)
                fnr = fdp_delta(fpr)
                return -fnr

            results = minimize_scalar(fun, method='Bounded', bounds=[0, 1-fpr],
                                      options={'disp': False})
            if results.success:
                return -results.fun
            else:
                return 0
    return fdp




def fdp_fdp_grad_to_approxdp(fdp, fdp_grad, log_flag = False):
    # when there is a dedicated implementation of fdp_grad

    # If the log flag is False, then
    #  fdp takes x \in [0,1] and output f(x)
    #  fdp_grad takes x \in [0,1] and output the subdifferential as an interval [grad_l, grad_h]


    # If log_flag is True, then it indicates that
    #   1. the first argument denotes log(1-fdp) as a function of logx
    #   2. the second argument denotes log(- partial fdp) as a function of logx

    if log_flag:
        fun1 = fdp
        fun2 = fdp_grad
    else:
        def fun1(logx):
            assert(logx <= 0)
            if np.isneginf(logx): # x == 0
                return np.log(1-fdp(0))
            elif logx == 0:  # x == 1
                return 1.0
            else:
                return np.log(1-fdp(np.exp(logx)))

        def fun2(logx):
            assert(logx <= 0)
            if np.isneginf(logx):
                grad_l, grad_h = fdp_grad(0)
            else:
                grad_l, grad_h = fdp_grad(np.exp(logx))
            log_neg_grad_l = np.log(-grad_l)
            log_neg_grad_h = np.log(-grad_h)

            if log_neg_grad_l > log_neg_grad_h:
                # in case the order is swapped
                tmp = log_neg_grad_h
                log_neg_grad_h = log_neg_grad_l
                log_neg_grad_l = tmp

            return log_neg_grad_l, log_neg_grad_h

    def find_logx(delta):
        def fun(logx):
            if np.isneginf(logx):
                output = np.log(delta) - fun1(logx)
                return output,output
            else:
                log_neg_grad_l, log_neg_grad_h = fun2(logx)
                log_one_minus_f = fun1(logx)
                low = utils.stable_logsumexp_two(log_neg_grad_l + logx,
                                                 np.log(delta)) - log_one_minus_f
                high = utils.stable_logsumexp_two(log_neg_grad_h + logx,
                                                  np.log(delta)) - log_one_minus_f
                return low, high

        def normal_equation(logx):
            if logx > 0:
                return np.inf
            low, high = fun(logx)
            if low <= 0 <= high:
                return 0
            else:
                return min(abs(high),abs(low))

        def normal_equation_loglogx(loglogx):
            logx = np.exp(loglogx)
            return normal_equation(logx)

        # find x such that y = 1-\delta
        tmp = fun1(np.log(1 - delta))
        if abs(tmp) < 1e-5:
            bound1 = np.log(-tmp - tmp**2 / 2 - tmp**3 / 6)
        else:
            bound1 = np.log(1-np.exp(fun1(np.log(1-delta))))
        #results = minimize_scalar(normal_equation, bracket=[-1,-2])
        results = minimize_scalar(normal_equation, method="Bounded", bounds=[bound1,0],
                                  options={'xatol': 1e-10, 'maxiter': 500, 'disp': 0})
        if results.success:
            if abs(results.fun) > 1e-4 and abs(results.x)>1e-10:
                # This means that we hit xatol (x is close to 0, but
                # the function value is not close to 0) In this case let's do an even larger search.
                raise RuntimeError("'find_logx' fails to find the tangent line.")
            else:
                return results.x
        else:
            raise RuntimeError(f"'find_logx' fails to find the tangent line: {results.message}")

    def approxdp(delta):
        if delta == 0:
            logx = -np.inf
            log_neg_grad_l, log_neg_grad_h = fun2(logx)
            return log_neg_grad_l
        elif delta == 1:
            return 0.0
        else:
            logx = find_logx(delta)
            log_one_minus_f = fun1(logx)
            # log_neg_grad_l, log_neg_grad_h = fun2(logx)
            s, mag = utils.stable_log_diff_exp(log_one_minus_f,np.log(delta))
            eps = mag - logx
            if eps < 0:
                return 0.0
            else:
                return eps

    #approxdp(1e-3)

    return approxdp



    # def findx(delta):
    #
    #     def fun(x):
    #         # if log_flag:
    #         #     if x == 0:
    #         #         return np.log(delta) - fun1(0)
    #         #     else:
    #         #         return utils.stable_logsumexp_two(fun2(x) + np.log(x), np.log(delta)) - fun1(x)
    #         #else:
    #         fx = fdp(x)
    #         if x == 0:
    #             output = np.log(delta) - np.log(1 - fx)
    #             return output
    #         else:
    #             grad_l, grad_h = fdp_grad(x)
    #         return utils.stable_logsumexp_two(np.log(-fdp_grad(x)) + np.log(x),
    #                                           np.log(delta)) - np.log(1-fx)
    #
    #     def normal_equation(x):
    #         return abs(fun(x))
    #     results = minimize_scalar(normal_equation, method="Bounded", bounds=[0,1],
    #                               options={'xatol': min(1e-10,1e-3*delta), 'maxiter': 500, 'disp': 0})
    #     if results.success:
    #         return results.x
    #     else:
    #         return None
    #
    # def approxdp(delta):
    #     x = findx(delta) - min(1e-10,1e-3*delta)
    #     if log_flag:
    #         return fun2(x)
    #     else:
    #         return np.log(-fdp_grad(x))
    # return approxdp






def cdf_to_approxdp(cdf_p,cdf_q, quadrature=True):
    """
    Solve eps(delta) query by searching a pair of cdf such that
    delta(eps) = 1 - cdf_p(eps) - e^eps*cdf_q(-eps)
    Args:
        cdf_p: the cdf of log(p/q)
        cdf_q: the cdf of log(q/p)
        quadrature: if False, using a FFT-based numerical tool to convert
        characteristic functions back to cdf.
    """
    if quadrature == False:
        # Future work: convert characteristic functions to cdf using FFT.
        return cdf_to_approxdp_fft(cdf_p,cdf_q, l=1e5)

    def trade_off(x):
        #x is e^epsilon
        log_e = np.log(x)
        # delta = 1 - result
        result = cdf_p(log_e) + x*cdf_q(-log_e)
        print('Binary search epsilon',log_e, 'current delta', 1-result)
        return result
    # tol denotes the relative difference in delta term
    exp_eps = numerical_inverse(trade_off, [0,1], tol=1e-2)
    def approxdp(delta):
        t = exp_eps(1 - delta)
        return np.log(t)
    return approxdp

def cdf_to_approxdp_fft(cdf_p,cdf_q, l=1e5):
    """
    Compute the numerical inversion of characteristic functions through FFT.

    Args:
        cdf_p: the cdf list of log(p/q)
        cdf_q: the cdf list of log(q/p)
        l: we compute epsilon(delta) for epsilon ranges from [-l, l]

    Return:
        an epsilon(delta) query
    """

    def approxdp(delta):
        t1 = time.time()
        cdf_p_list = cdf_p(l)
        cdf_q_list = cdf_q(l)
        t2 = time.time()
        print('time in computing cdf', t2-t1)
        # H is the number of point in fft, H = 2N-1
        H = len(cdf_p_list)
        mesh_size = 2.*l/H
        n = int((H-1)/2)
        b = l
        f = lambda x: -b + mesh_size * x
        eps_list= [f(i) for i in range(H)]
        i = n-1
        j = H-1
        # Binary search to search a suitable epsilon.
        t1 = time.time()
        while i<j:
            mid = int((i+j)/2)
            eps = eps_list[mid]
            result = cdf_p_list[int(mid)] + np.exp(eps) * cdf_q_list[int(H-mid-4)]
            if abs(1-result)<0.1*delta:
                return eps_list[mid]
            if (1-result)>delta:
                i = mid+1
            else:
                j = mid-1
        t2 = time.time()
        #print('time used in binary search', t2-t1)
        if i<H-1 and eps_list[j]>0:
            return eps_list[j]
        raise ValueError("Mesh size is too large or l is smaller than the range of epsilon,"
                         " please check the parameters N and l.")


    return approxdp


def pdf_to_phi(p, q, t):
    """
    Given a pair of dominating pair distribution p and q, returns the log phi function of log(p/q) and loq(q/p).

    Args:
        p,q: the pdf of dominating pair distribution
    """
    # compute the log_phi np.exp(1.j *t*log(p(o)/q(o)))
    def _inte(y):
        o = y*1.0/(1-y**2)
        s = np.exp(1.0j * np.log(p(o)/q(o)) * t)*p(o)
        # Improve the stability (NaN is close to zero).
        s[np.isnan(s)] = 0
        return s
    # exp_term = lambda o: np.exp(1.0j * np.log(p(o)/q(o)) * t)*p(o)
    inte_f_p = lambda y: _inte(y) * (1 + y ** 2) / ((1 - y ** 2) ** 2)
    # applies gaussian quadrature to compute the expectation
    res_p = integrate.quadrature(inte_f_p, -1.0, 1.0, tol=1e-15, rtol=1e-15, maxiter=100)
    return np.log(res_p[0])


def phi_to_cdf(log_phi, ell, n_quad=500, extra_para=None):
    """
     This function computes the CDF of privacy loss R.V. via Levy theorem.
     https://en.wikipedia.org/wiki/Characteristic_function_%28probability_theory%29#Inversion_formulae
     The integration is implemented through Gaussian quadrature.

     Args:
        log_phi: the log of characteristic (phi)  function.
        ell: the privacy loss RV is evaluated at ell
        extra_para: extra parameters used to describe the privacy loss R.V..

    Return: the CDF of the privacy loss RV when evaluated at ell。
    """

    def qua(t):
        """
        Convert [-1, 1] to an infinite integral.
        """
        new_t = t*1.0/(1-t**2)
        phi_result = [log_phi(x) for x in new_t]
        inte_function = 1.j/new_t * np.exp(-1.j*new_t*ell)*np.exp(phi_result)
        return inte_function
    # n is the maximum sampling point used in Gaussian quadrature, setting it to be >700 is usually very accurate.
    inte_f = lambda t: qua(t) * (1 + t ** 2) / ((1 - t ** 2) ** 2)
    res = integrate.fixed_quad(inte_f, -1.0, 1.0, n =n_quad)

    result = res[0]
    return np.real(result)/(2*np.pi)+0.5


def cdf_approx_fft(log_phi, L, N=5e6):
    """
     Future work: This function converts the characteristic function to the CDF using FFT.

     The detailed numerical inversion can be found in Algorithm B in
     https://www.tandfonline.com/doi/abs/10.1080/03461238.1975.10405087. This approach
     does not involve truncation errors.
     We consider privacy loss R.V. is zj: = -b + lam*j, where lam = 2pi/(eta*(2N-1)).


     Args:
        log_phi: the log of characteristic (phi)  function.
        L: Limit for the approximation of the privacy loss distribution integral.
        N: Number of points in FFT (FFT is over 2N-1 points)

    Return:
        A list of CDFs of privacy loss R.V. evaluated between [-L, L]

    """

    # evaluate the j-th privacy loss r.v. zj: = -b + lam*j, where lam = 2pi/(eta*(2N-1)).
    eta = np.pi / L
    N = int(N)
    b = -np.pi / eta
    lam = 2 * np.pi / (eta * (2 * N - 1))
    t_list = [m_hat + 1 - N for m_hat in range(2 * N - 1)]
    t0 = time.time()
    c_nu = lambda t: (1 - t) * np.cos(np.pi * t) + np.sin(np.pi * t) / np.pi

    # FFT is used for the calculation of the trigonometric sums appearing in the formulas.
    # l == 0 is undefined in the original formula, thus we need to exclude it from the fft.
    def f_phi(l):
        if l == 0:
            return 0
        nu = l * 1.0 / N
        c_t = c_nu(abs(nu))
        return c_t * np.exp(log_phi(l * eta) - 1.j * eta * b * l) / l

    phi_list = [f_phi(m_hat) for m_hat in t_list]
    fft_res = fft(phi_list)

    fft_norm = [fft_res[j] * np.exp(1.j * (N - 1) * 2.0 / (2. * N - 1) * j * np.pi) for j in range(2 * N - 1)]

    """
    cdf[j] denotes the cdf of log(p/q) when evaluates at zj, zj: = -b + lam*j.
    the range of z is [-pi/eta, pi/eta], the mesh-size on zj is lam. To get a good approximation on cdf,
     we  need eta*N to be a large number.

    cdf(z) = 0.5 + eta*z/(2pi) -fft_res[z]
    """
    convert_z = lambda j: b + lam * j
    cdf = [0.5 + eta * convert_z(j) / (2 * np.pi) - 1. / (2 * np.pi * 1.j) * fft_norm[j] for j in range(2 * N - 1)]
    return cdf


def cdf_to_approxdelta_fft(cdf_p, cdf_q, l =1e4):
    """
    Future work: Returns delta as a function of epsilon using FFT.

    Args:
        cdf_p: a list cdf of log(p/q)
        cdf_q: the cdf of log(q/p)
    """

    def approx_delta(eps):
        cdf_p_list = cdf_p(l)
        cdf_q_list = cdf_q(l)
        if eps > l:
            raise ValueError("Epsilon out of [-L,L] window, please check the parameters.")
        H = len(cdf_p_list)
        mesh_size = 2. * l / H
        b = mesh_size * H / 2
        f = lambda x: -b + mesh_size * x
        eps_list = [f(i) for i in range(H)]
        # idx is the first idx where eps_list[idx]> eps
        idx = np.where(eps_list>eps)[0][0]
        delta = 1-cdf_p_list[idx] - np.exp(eps_list[idx]) * cdf_q_list[int(H-idx-1)]
        return delta
    return approx_delta


def cdf_to_approxdelta(cdf_p, cdf_q, quadrature=True):
    """
    Returns delta as a function of epsilon.

    Args:
        cdf_p: the cdf of log(p/q)
        cdf_q: the cdf of log(q/p)
        quadrature: if True, apply Gaussian quadrature for numerical inversion (converse
        characteristic functions back to CDFs). Else, using FFT instead.

    """
    if quadrature == False:
        return cdf_to_approxdelta_fft(cdf_p, cdf_q)

    def approx_delta(eps):

        delta = 1-cdf_p(eps) - np.exp(eps) * cdf_q(-eps)
        return delta
    return approx_delta



def fdp_to_approxdp(fdp):
    # Check out Proposition 2.12 of Dong, Roth and Su
    # if given a symmetric fdp function f,
    # its convex conjugate with some manipulation defines the \delta as a function of \epsilon.
    # How to calculate convex conjugates? Can we just do numerical computation?
    # How to ensure the symmetry of fdp function f?
    # One way we can go is to define an fdp_bank where we code up the conjugate pairs in analytical form
    # This allows one to more easily convert approxdp function to fdp

    fstar = conjugate(fdp)
    # def delta_from_fdp(eps):
    #     return 1 + fstar(-np.exp(eps))
    #
    # approxdp = numerical_inverse(delta_from_fdp)

    # def neg_log_one_plus_fstar_neg_input(x):
    #     return -np.log(1 + fstar(-x))
    #
    # exp_eps = numerical_inverse(neg_log_one_plus_fstar_neg_input)
    # def approxdp(delta):
    #     return np.log(exp_eps(-np.log(delta)))

    def neg_fstar_neg_input(x):
        return -1.0*fstar(-1.0*x)

    exp_eps = numerical_inverse(neg_fstar_neg_input,[0,1])
    def approxdp(delta):
        return np.log(exp_eps(1-delta))

    return approxdp


def numerical_inverse(f, bounds=[1, np.inf], tol=1e-6):
    # of a scalar, monotonic function
    def inv_f(y):
        if bounds:
            if y > bounds[1] or y < bounds[0]:
                return None

        def fun(x):
            return f(x) - y

        # The domain should be encoded in the definition of f directly.
        def normal_equation(x):
            return abs(fun(x))

        #results = minimize_scalar(normal_equation, bounds=bounds, bracket=[1, 2], tol=1e-10)
        results = minimize_scalar(normal_equation, bounds=bounds, bracket=[1,2], tol=tol)


        #results = root_scalar(fun, options={'disp': False})
        if results.success:
            return results.x
        else:
            return None

    return inv_f


def approxdp_from_its_inverse(delta_func):
    # Convert delta as a function of epsilon to epsilon as a function of delta
    return numerical_inverse(delta_func, bounds=[0,1])



## Utility functions

def conjugate(f,tol=1e-10):
    # numerically evaluate convex conjugate of a convex function f: [0,1] --> [0,1]
    # domain of the function y is [0,1]
    def fstar(x):
        def fun(y):
            return -y*x + f(y)
        results = minimize_scalar(fun, method='Bounded', bounds=(0, 1),
                                  options={'disp': False,'xatol':tol})
        if results.success:
            return -(results.fun + tol)
            # output an upper bound
        else:
            return None
    return fstar


def pointwise_minimum(f1, f2):
    def min_f1_f2(x):
        return np.minimum(f1(x), f2(x))
    return min_f1_f2

def pointwise_minimum_two_arguments(f1, f2):
    def min_f1_f2(x, y):
        return np.minimum(f1(x, y), f2(x, y))
    return min_f1_f2

def pointwise_maximum(f1, f2):
    def max_f1_f2(x):
        return np.maximum(f1(x), f2(x))
    return max_f1_f2



"""
'calibrator_zoo' module implements a number of ways to choose parameters of a mechanism to achieve
a pre-defined privacy guarantee.

For instance, we could calibrate noise to sensitivity in Laplace mechanism to achieve any given \eps

All calibrators inherit the autodp_core.calibrator

"""

from math import exp, sqrt
from scipy.special import erf
from autodp.autodp_core import Calibrator
from scipy.optimize import minimize_scalar, root_scalar

class eps_delta_calibrator(Calibrator):
    """
    Noting there exists multiple parameters in mechanisms. we must specify which parameter requires optimization.
    """
    def __init__(self):
        Calibrator.__init__(self)
        self.name = 'eps_delta_calibrator'

        # Update the function that is callable

        self.calibrate = self.param_from_eps_delta

    def param_from_eps_delta(self,mech_class,eps,delta, bounds,name=None):
        def get_eps(x,delta):

            mech = mech_class(x)
            return mech.get_approxDP(delta)

        def err(x):
            return abs(eps-get_eps(x,delta))

        results = minimize_scalar(err, method='bounded', bounds=bounds)
        if results.success and results.fun < 1e-3:
            if name:
                mech = mech_class(results.x,name=name)
            else:
                mech = mech_class(results.x)
            return mech
        else:
            raise RuntimeError(f"eps_delta_calibrator fails to find a parameter: {results.message}")


class generalized_eps_delta_calibrator(Calibrator):
    """
    Noting there exists multiple parameters in mechanisms. we must specify which parameter requires optimization.
    """
    def __init__(self):
        Calibrator.__init__(self)
        self.name = 'generalized_eps_delta_calibrator'

        # Update the function that is callable

        self.calibrate = self.param_from_eps_delta

    def param_from_eps_delta(self,mech_class,eps,delta, bounds, params=None, para_name=None, name=None):
        """
        params defines all parameters, include those need to optimize over
        para_name is the parameter that we need to tune

        for example, in subsampleGaussian mechanism
        params = {'prob':prob, 'sigma':sigma, 'coeff':coeff}, where coeff and sigma are given as the fixed input
        """
        def get_eps(x,delta):
            if params is None:
                # params is None implies the mechanism only has one parameter
                mech = mech_class(x)
            else:
                params[para_name] = x
                mech = mech_class(params)
            return mech.get_approxDP(delta)

        def err(x):
            return abs(eps-get_eps(x,delta))

        results = minimize_scalar(err, method='bounded', bounds=bounds)
        if results.success and results.fun < 1e-3:

            if params is None:
                mech = mech_class(results.x, name=name)
            else:
                params[para_name] = results.x
                mech = mech_class(params,name=name)
            return mech
        else:
            raise RuntimeError(f"eps_delta_calibrator fails to find a parameter: {results.message}")



class ana_gaussian_calibrator(Calibrator):
    """ Calibrate a Gaussian perturbation for differential privacy using the analytic Gaussian mechanism of [Balle and Wang, ICML'18]

    Modified from https://github.com/BorjaBalle/analytic-gaussian-mechanism/blob/master/agm-example.py

    Arguments:
    epsilon : target epsilon (epsilon > 0)
    delta : target delta (0 < delta < 1)
    tol : error tolerance for binary search (tol > 0)
    Output:
    params : a dictionary that contains field `sigma' --- the standard deviation of Gaussian noise needed to achieve
        (epsilon,delta)-DP under global sensitivity 1

    Note that this one does not support composition or subsample
    """


    def __init__(self):
        Calibrator.__init__(self)
        self.name = 'ana_gaussian_calibrator'

        # Update the function that is callable

        self.calibrate = self.param_from_eps_delta

    def param_from_eps_delta(self,mech_class,eps,delta,name=None, tol=1.e-12):
        def Phi(t):
            return 0.5 * (1.0 + erf(float(t) / sqrt(2.0)))

        def caseA(epsilon, s):
            return Phi(sqrt(epsilon * s)) - exp(epsilon) * Phi(-sqrt(epsilon * (s + 2.0)))

        def caseB(epsilon, s):
            return Phi(-sqrt(epsilon * s)) - exp(epsilon) * Phi(-sqrt(epsilon * (s + 2.0)))

        def doubling_trick(predicate_stop, s_inf, s_sup):
            while (not predicate_stop(s_sup)):
                s_inf = s_sup
                s_sup = 2.0 * s_inf
            return s_inf, s_sup

        def binary_search(predicate_stop, predicate_left, s_inf, s_sup):
            s_mid = s_inf + (s_sup - s_inf) / 2.0
            while (not predicate_stop(s_mid)):
                if (predicate_left(s_mid)):
                    s_sup = s_mid
                else:
                    s_inf = s_mid
                s_mid = s_inf + (s_sup - s_inf) / 2.0
            return s_mid

        delta_thr = caseA(eps, 0.0)

        if (delta == delta_thr):
            alpha = 1.0
            sigma = alpha / sqrt(2.0 * eps)

        else:
            if (delta > delta_thr):
                predicate_stop_DT = lambda s: caseA(eps, s) >= delta
                function_s_to_delta = lambda s: caseA(eps, s)
                predicate_left_BS = lambda s: function_s_to_delta(s) > delta
                function_s_to_alpha = lambda s: sqrt(1.0 + s / 2.0) - sqrt(s / 2.0)

            else:
                predicate_stop_DT = lambda s: caseB(eps, s) <= delta
                function_s_to_delta = lambda s: caseB(eps, s)
                predicate_left_BS = lambda s: function_s_to_delta(s) < delta
                function_s_to_alpha = lambda s: sqrt(1.0 + s / 2.0) + sqrt(s / 2.0)

            predicate_stop_BS = lambda s: abs(function_s_to_delta(s) - delta) <= tol

            s_inf, s_sup = doubling_trick(predicate_stop_DT, 0.0, 1.0)
            s_final = binary_search(predicate_stop_BS, predicate_left_BS, s_inf, s_sup)
            alpha = function_s_to_alpha(s_final)
            sigma = alpha / sqrt(2.0 * eps)


        if name:
            mech = mech_class(sigma,name=name)
        else:
            mech = mech_class(sigma)
        return mech



#TODO: implement other generic and specialized privacycalibrators.
#TODO: deprecate the API from "privacy_calibrator.py"




# This module implements f-DP function templates for various mechanisms
# Note that f-DP is equivalent to the family of (eps,delta)-DP described by the functions
#  delta(eps) or eps(delta), generic conversion is available from the converter module
# however it is numerically better if we implement and support cases when fdp is natural

import numpy as np
from scipy.stats import norm
from autodp import utils


def fDP_gaussian(params, fpr):
    """
    :param params:
        'sigma' --- is the normalized noise level: std divided by global L2 sensitivity
    :param fpr: False positive rate --- input to the fDP function
    :return: Evaluation of the fnr lower bound supported by the gaussian mechanism
    """
    sigma = params['sigma']
    # assert(sigma > 0)
    assert(sigma >= 0)
    if sigma == 0:
        return 0
    else:
        return norm.cdf(norm.ppf(1-fpr)-1/sigma)


def fDP_approx_DP(params, fpr):
    """
    :param params:
        'eps' --- DP parameter 1
        'delta' --- DP parameter 2
    :param fpr: False positive rate --- input to the fDP function
    :return:  Evaluation of the fnr lower bound supported by any pure DP mechanism
    """
    eps = params['eps']
    delta = params['delta']
    assert(eps >= 0)
    if np.isinf(eps):
        return 0
    elif eps == 0:
        return 1-fpr - delta
    else:
        return np.max([0, 1-delta - np.exp(eps)*fpr, np.exp(-eps)(1-delta - fpr)])


def fDP_pure_DP(params, fpr):
    """
    :param params:
        'eps' --- DP parameter
    :param fpr: False positive rate --- input to the fDP function
    :return:  Evaluation of the fnr lower bound supported by any pure DP mechanism
    """
    params['delta'] = 0
    return fDP_approx_DP(params, fpr)


def fdp_grad_gaussian(params,fpr):
    """
    :param params:
        'sigma' --- is the normalized noise level: std divided by global L2 sensitivity
    :param fpr: False positive rate --- input to the fDP function
    :return: Evaluation of derivative of the Tradeoff function at input fpr
    """
    sigma = params['sigma']
    # assert(sigma > 0)
    assert(sigma >= 0)
    if sigma == 0:
        return 0
    else:
        return -norm.pdf(norm.ppf(1-fpr)-1/sigma)/norm.pdf(norm.ppf(1-fpr))

def log_one_minus_fdp_gaussian(params, logfpr):
    """
    :param params:
        'sigma' --- is the normalized noise level: std divided by global L2 sensitivity
    :param logfpr: log of False positive rate --- input to the fDP function
    :return: log(1-f(x)).
    """
    sigma = params['sigma']
    # assert(sigma > 0)
    assert(sigma >= 0)
    if sigma == 0:
        return 0
    else:
        if np.isneginf(logfpr):
            return -np.inf
        else:

            norm_ppf_one_minus_fpr = utils.stable_norm_ppf_one_minus_x(logfpr)

            return norm.logsf(norm_ppf_one_minus_fpr-1/sigma)




def log_neg_fdp_grad_gaussian(params, logfpr):
    """
    :param params:
        'sigma' --- is the normalized noise level: std divided by global L2 sensitivity
    :param logfpr: log of False positive rate --- input to the fDP function
    :return: log(-partial f(x))
    """
    sigma = params['sigma']
    # assert(sigma > 0)
    assert(sigma >= 0)
    if sigma == 0:
        return 0
    else:
        if np.isneginf(logfpr):  # == 0:
            return np.inf, np.inf
        elif logfpr == 0:  #fpr == 1:
            return -np.inf, -np.inf
        else:
            norm_ppf_one_minus_fpr = utils.stable_norm_ppf_one_minus_x(logfpr)
            grad = -(norm_ppf_one_minus_fpr
                     - 1 / sigma) ** 2 / 2 + norm_ppf_one_minus_fpr ** 2 / 2
            return grad, grad
